"""Class which tests the AST utils class."""
import astroid

from dslinter.utils.ast import ASTUtil


class TestAST:
    """Class which tests the AST utils class."""

    def test_search_nodes(self):
        """Test the search_nodes method."""
        module_tree = astroid.parse(
            """
            a = b.c(d)

            def e(f):
                return g.h(i)
            """
        )
        found = ASTUtil.search_nodes(module_tree, astroid.Call)
        # noinspection PyUnresolvedReferences
        assert len(found) == 2 and found[0].func.attrname == "c" and found[1].func.attrname == "h"

    def test_get_source_code(self):
        """Test the get_source_code method."""
        source_code = "a = b.c(d)"
        module_tree = astroid.parse(source_code)
        assert ASTUtil.get_source_code(module_tree) == source_code

    def test_search_body_parent_module(self):
        """Test whether the module is returned when searching for the parent of its child."""
        module_tree = astroid.parse(
            """
            f()
            """
        )
        node = module_tree.body[0]
        assert ASTUtil.search_body_parent(node) == module_tree

    def test_search_body(self):
        """Test whether the correct body is returned."""
        module_tree = astroid.parse(
            """
            f()
            """
        )
        node = module_tree.body[0]
        assert ASTUtil.search_body(node) == module_tree.body

    def test_search_body_parent_function(self):
        """Test whether the function is returned when searching for the parent of its child."""
        module_tree = astroid.parse(
            """
            def f():
                return 0
            """
        )
        node = module_tree.body[0].body[0]
        assert ASTUtil.search_body_parent(node) == module_tree.body[0]



"""Checker which checks whether tf.TensorArray() is used for growing array in the loop in tensorflow code. """
import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker

from dslinter.utils.exception_handler import ExceptionHandler
from dslinter.utils.type_inference import TypeInference
from typing import Dict


class TensorArrayTensorflowChecker(BaseChecker):
    """Checker which checks whether tf.TensorArray() is used for growing array in the loop in tensorflow code. """

    __implements__ = IAstroidChecker

    name = "tensor-array-tensorflow"
    priority = -1
    msgs = {
        "W5515": (
            "The tf.constant() variable is assigned or growing in the loop.",
            "tensor-array-tensorflow",
            "Use tf.TensorArray() for growing array in the loop.",
        )
    }
    options = ()

     # [variable name, inferred type of object the function is called on]
    _variable_types: Dict[str, str] = {}

    def visit_module(self, module: astroid.Module):
        """Visit module and infer which libraries the variables are from."""
        try:
            self._variable_types = TypeInference.infer_library_variable_first_types(module)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, module)

    def visit_for(self, for_node: astroid.For):
        """Visit for node and see whether the rule is violated."""
        try:
            for node in for_node.body:
                if(
                    isinstance(node, astroid.Assign)
                    and len(node.targets) > 0
                    and hasattr(node.targets[0], "name")
                    and node.targets[0].name in self._variable_types
                    and self._variable_types[node.targets[0].name] in ["tf.constant", "tensorflow.constant"]
                    and self.infer_call_expression(node.value) in ["tf.concat"]
                ):
                    self.add_message("tensor-array-tensorflow", node=for_node)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, for_node)

    @staticmethod
    def infer_call_expression(call_node: astroid.Call):
        """Infer the whole call exprassion of the call node."""
        call_expression = ""
        call = call_node.func
        if hasattr(call, "attrname"):
            call_expression = "." + call.attrname + call_expression
        while hasattr(call, "expr"):
            call = call.expr
            if hasattr(call, "attrname"):
                call_expression = "." + call.attrname + call_expression
        if hasattr(call, "name"):
            call_expression = call.name + call_expression
        return call_expression



"""Class which tests DataframeConversionPandasChecker."""
import astroid
import pylint.testutils

import dslinter


class TestDataframeConversionPandasChecker(pylint.testutils.CheckerTestCase):
    """Class which tests DataframeConversionPandasChecker."""

    CHECKER_CLASS = dslinter.plugin.DataframeConversionPandasChecker

    def test_dataframe_conversion_incorrectly_used1(self):
        """Message should be added if .values is used for dataframe conversion."""
        script = """
        import numpy as np
        import pandas as pd #@
        
        index = [1, 2, 3, 4, 5, 6, 7]
        a = [np.nan, np.nan, np.nan, 0.1, 0.1, 0.1, 0.1]
        b = [0.2, np.nan, 0.2, 0.2, 0.2, np.nan, np.nan]
        c = [np.nan, 0.5, 0.5, np.nan, 0.5, 0.5, np.nan]
        df = pd.DataFrame({'A': a, 'B': b, 'C': c}, index=index)
        # df = df.rename_axis('ID')
        arr = df.values #@     
        """
        import_node, assign_node = astroid.extract_node(script)
        call_node = assign_node.value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="dataframe-conversion-pandas", node=call_node)):
            self.checker.visit_import(import_node)
            self.checker.visit_call(call_node)

    def test_dataframe_conversion_correctly_used(self):
        """No message should be added if .to_numpy() is used for dataframe conversion."""
        script = """
        import numpy as np
        import pandas as pd #@
        
        index = [1, 2, 3, 4, 5, 6, 7]
        a = [np.nan, np.nan, np.nan, 0.1, 0.1, 0.1, 0.1]
        b = [0.2, np.nan, 0.2, 0.2, 0.2, np.nan, np.nan]
        c = [np.nan, 0.5, 0.5, np.nan, 0.5, 0.5, np.nan]
        df = pd.DataFrame({'A': a, 'B': b, 'C': c}, index=index)
        # df = df.rename_axis('ID')
        arr = df.to_numpy()   #@
        """
        import_node, assign_node = astroid.extract_node(script)
        call_node = assign_node.value
        with self.assertNoMessages():
            self.checker.visit_import(import_node)
            self.checker.visit_call(call_node)

    def test_dataframe_conversion_incorrectly_used2(self):
        """Message should be added if .values is used for dataframe conversion."""
        script = """
        # LOAD LIBRARIES
        import pandas as pd #@
        import numpy as np
        from sklearn.model_selection import train_test_split
        from keras.utils.np_utils import to_categorical
        from keras.models import Sequential
        from keras.layers import Dense, Dropout, Flatten, Conv2D, MaxPool2D, BatchNormalization
        from keras.preprocessing.image import ImageDataGenerator
        from keras.callbacks import LearningRateScheduler

        # # Load Kaggle's 42,000 training images

        # In[ ]:

        # LOAD THE DATA
        train = pd.read_csv("../input/train.csv")
        test = pd.read_csv("../input/test.csv")

        # In[ ]:

        # PREPARE DATA FOR NEURAL NETWORK
        Y_train = train["label"]
        X_train = train.drop(labels = ["label"],axis = 1)
        X_train = X_train / 255.0
        X_test = test / 255.0
        X_train = X_train.values.reshape(-1,28,28,1) #@
        X_test = X_test.values.reshape(-1,28,28,1)
        Y_train = to_categorical(Y_train, num_classes = 10)
        """
        import_node, assign_node = astroid.extract_node(script)
        call_node = assign_node.value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="dataframe-conversion-pandas", node=call_node)):
            self.checker.visit_import(import_node)
            self.checker.visit_call(call_node)

    def test_dataframe_conversion_incorrectly_used3(self):
        """Message should be added if .values is used for dataframe conversion."""
        script = """
        import pandas as pd #@
        price_mm = MinMaxScaler().fit_transform(price.values.reshape(-1, 1).astype(np.float64)).flatten() #@
        """
        import_node, assign_node = astroid.extract_node(script)
        call_node = assign_node.value.func.expr.args[0].func.expr
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="dataframe-conversion-pandas", node=call_node)):
            self.checker.visit_import(import_node)
            self.checker.visit_call(call_node)



"""Checker which check whether df.values is used for dataframe conversion."""
from typing import Dict
import astroid
from pylint.interfaces import IAstroidChecker
from pylint.checkers import BaseChecker

from dslinter.utils.exception_handler import ExceptionHandler


class DataframeConversionPandasChecker(BaseChecker):
    """Checker which check whether df.values is used for dataframe conversion."""

    __implements__ = IAstroidChecker

    name = "dataframe-conversion-pandas"
    priority = -1
    msgs = {
        "W5504": (
            "df.values is used in pandas code for dataframe conversion.",
            "dataframe-conversion-pandas",
            "For dataframe conversion in pandas code, use .to_numpy() instead of .values."
        )
    }
    options = ()

    # [variable name, inferred type of object the function is called on]
    _call_types: Dict[str, str] = {}
    _imported_pandas = False

    def visit_import(self, import_node: astroid.Import):
        """Visit import node to see whether pandas is imported."""
        try:
            for name, _ in import_node.names:
                if name == "pandas":
                    self._imported_pandas = True
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, import_node)

    def visit_call(self, call_node: astroid.Call):
        """Visit call node to see whether there is rule violation."""
        try:
            if hasattr(call_node, "attrname") and call_node.attrname == "values":
                self.add_message("dataframe-conversion-pandas", node=call_node)
                return
            if(
                hasattr(call_node, "func")
                and hasattr(call_node.func, "expr")
                and hasattr(call_node.func.expr, "attrname")
                and call_node.func.expr.attrname == "values"
            ):
                    self.add_message("dataframe-conversion-pandas", node=call_node)
                    return
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, call_node)



"""Check whether there is an unnecessary iteration in Tensorflow code."""
from typing import Dict
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker
import astroid
from dslinter.utils.exception_handler import ExceptionHandler
from dslinter.utils.type_inference import TypeInference


class UnnecessaryIterationTensorflowChecker(BaseChecker):
    """Check whether there is an unnecessary iteration in Tensorflow code."""
    __implements__ = IAstroidChecker

    name = "unnecessary-iteration-tensorflow"
    priority = -1
    msgs = {
        "R5502": {
            "There is an unnecessary iteration in the tensorflow code.",
            "unnecessary-iteration-tensorflow",
            "There is a efficient solution(Vectorization or Reduction) to replace the iteration.",
        }
    }
    options = ()

    # [variable name, inferred type of object the function is called on]
    _variable_types: Dict[str, str] = {}

    def visit_module(self, module: astroid.Module):
        """Visit module and infer which library the variables are from. """
        try:
            self._variable_types = TypeInference.infer_native_variable_most_recent_types(module)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, module)

    def visit_for(self, node: astroid.For):
        """Evaluate whether there is an augmented assign in the loop, it can be replaced
            by a reduction operation, which is faster."""
        try:
            if (
                hasattr(node, "body")
                # there is augmented assign with tf variable in the body
                and self._augmented_assign_with_tf_variable(node.body)
            ):
                self.add_message("iteration-tensorflow", node=node)
        except:  # pylint: disable=W0702
            ExceptionHandler.handle(self, node)

    def _augmented_assign_with_tf_variable(self, body: list) -> bool:
        """
        Check whether there are augmented assigns in the body,
        and whether the augmented assign is applied on a tensorflow variable.
        :param body: body in the for node
        :return: True when meeting the requirements
        """
        for node in body:
            # there is augmented assign in the body
            if isinstance(node, astroid.AugAssign):
                # # the augmented assign is apply on a const
                # if hasattr(node, "value") and isinstance(node.value, astroid.nodes.Const):
                #     return True
                while hasattr(node, "value"):
                    node = node.value
                # the augmented assign is apply on a tf variable node
                if hasattr(node, "name") and self._is_tf_variable(node.name):
                    return True
        return False

    def _is_tf_variable(self, name) -> bool:
        """
        Check whether the variable with the name is a tensorflow variable
        :param name: name of the variable
        :return: True when meeting the requirements
        """
        return name in self._variable_types and self._variable_types[name] in ["tf", "tensorflow"]



"""Class which tests the InPlaceNumpyChecker"""
import astroid
import pylint.testutils
import dslinter


class TestInPlaceNumpyChecker(pylint.testutils.CheckerTestCase):
    """Class which tests the InPlaceNumpyChecker"""

    CHECKER_CLASS = dslinter.plugin.InPlaceNumpyChecker

    def test_operation_call_unassigned1(self):
        """Test whether a message is added when the result is lost."""
        script = """
        import numpy as np
        np.clip([1,2,3] ,1, -1) #@
        """
        unassigned_call = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="inplace-numpy", node=unassigned_call)):
            self.checker.visit_call(unassigned_call)

    def test_operation_call_unassigned2(self):
        """Test whether a message is added when the result is lost."""
        script = """
        import numpy as np
        np.sin(1) #@
        """
        unassigned_call = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="inplace-numpy", node=unassigned_call)):
            self.checker.visit_call(unassigned_call)

    def test_operation_call_assigned(self):
        """Test whether no message is added when the result is assigned."""
        script = """
        import numpy as np
        a = np.clip([1,2,3], 1, -1) #@
        """
        assigned_call = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_call(assigned_call)

    def test_operation_call_inplace(self):
        """Test whether no message is added when the inplace parameter is set."""
        script = """
        import numpy as np
        a = np.array([1,2,3]) 
        np.clip(a, 1, -1, out = a) #@
        """
        unassigned_call = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_call(unassigned_call)

    def test_np_call_in_a_function(self):
        """Test whether no message is added when the call is in a another call."""
        script = """
        import numpy as np
        print(np.sum(y)) #@
        """
        call_node = astroid.extract_node(script).args[0]
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    def test_np_call_in_return_node(self):
        """Test whether no message is added when the call is in a return node."""
        script = """
        import numpy as np
        def test_function():
            return np.sum() #@
        """
        call_node = astroid.extract_node(script).value
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    def test_call_name_in_whitelist(self):
        """Tesr whether no message is added if the call name is in whitelist"""
        script = """
        import numpy as np
        a = [1, 2, 3]
        np.save(a) #@
        """
        call_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_call(call_node)



"""Checker which checks whether self.net() is used to forward the input into the network in PyTorch instead of self.net.forward()."""
import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker

from dslinter.utils.exception_handler import ExceptionHandler


class ForwardPytorchChecker(BaseChecker):
    """Checker which checks whether self.net() is used to forward the input into the network in PyTorch instead of self.net.forward()."""

    __implements__ = IAstroidChecker

    name = "forward-pytorch"
    priority = -1
    msgs = {
        "W5516": (
            "The self.net.forward() is used in the code rather than self.net().",
            "forward-pytorch",
            "It is recommended to use self.net() rather than self.net.forward() in PyTorch code."
        )
    }
    options = ()

    def visit_call(self, call_node: astroid.Call):
        """
        When a Call node is visited, check whether it violated the rule in this checker.
        :param call_node: The node which is visited.
        """
        try:
            _has_forward = False
            _call_from_self = False
            _call_from_super = False
            if hasattr(call_node.func, "attrname") and call_node.func.attrname == "forward":
                _has_forward = True
            if(
                hasattr(call_node.func, "expr")
                and hasattr(call_node.func.expr, "name")
                and call_node.func.expr.name == "self"
            ):
                _call_from_self = True
            if(
                hasattr(call_node.func, "expr")
                and hasattr(call_node.func.expr, "func")
                and hasattr(call_node.func.expr.func, "name")
                and call_node.func.expr.func.name == "super"
            ):
                _call_from_super = True
            if _has_forward is True and (_call_from_self is False and _call_from_super is False):
                self.add_message("forward-pytorch", node=call_node)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, call_node)



"""Class which tests the HyperparameterPyTorchChecker."""
import astroid
import pylint.testutils
import dslinter.plugin


class TestHyperparameterPyTorchChecker(pylint.testutils.CheckerTestCase):
    """Class which tests the HyperparameterPyTorchChecker."""

    CHECKER_CLASS = dslinter.plugin.HyperparameterPyTorchChecker

    def test_batch_size_set(self):
        script = """
        from torch.utils.data import DataLoader #@
        DataLoader(dataset, batch_size=4)     #@
        """
        importfrom_node, call_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_importfrom(importfrom_node)
            self.checker.visit_call(call_node)

    def test_batch_size_not_set(self):
        script = """
        from torch.utils.data import DataLoader #@
        DataLoader(dataset)   #@
        """
        importfrom_node, call_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="hyperparameters-pytorch", node=call_node)):
            self.checker.visit_importfrom(importfrom_node)
            self.checker.visit_call(call_node)

    def test_momentum_set(self):
        script = """
        from torch.optim import SGD #@
        optimizer = SGD(model.parameters(), lr=0.01, momentum=0.9, weight_decay = 0) #@
        """
        importfrom_node, assign_node = astroid.extract_node(script)
        call_node = assign_node.value
        with self.assertNoMessages():
            self.checker.visit_importfrom(importfrom_node)
            self.checker.visit_call(call_node)

    def test_momentum_not_set(self):
        script = """
        from torch.optim import SGD #@
        optimizer = SGD(model.parameters(), lr=0.01) #@
        """
        importfrom_node, assign_node = astroid.extract_node(script)
        call_node = assign_node.value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="hyperparameters-pytorch", node=call_node)):
            self.checker.visit_importfrom(importfrom_node)
            self.checker.visit_call(call_node)



"""Checker which checks whether chain indexing is used in pandas code."""
import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker
from typing import Dict

from dslinter.utils.exception_handler import ExceptionHandler
from dslinter.utils.type_inference import TypeInference


class ChainIndexingPandasChecker(BaseChecker):
    """Checker which checks whether chain indexing is used in pandas code."""

    __implements__ = IAstroidChecker

    name = "chain-indexing-pandas"
    priority = -1
    msgs = {
        "W5502": (
            "Chain indexing is used in pandas code.",
            "chain-indexing-pandas",
            "Chain indexing is considered bad practice in pandas code and should be avoided."
        )
    }
    options = ()

    # [subscript node name, inferred type of object the function is called on]
    _subscript_types: Dict[str, str] = {}

    def visit_module(self, module: astroid.Module):
        """Visit module and infer which libraries the variables are from. """
        try:
            self._subscript_types = TypeInference.infer_library_variable_most_recent_types(module)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, module)

    def visit_subscript(self, subscript_node: astroid.Subscript):
        """Visit subscript node and check whether there is chain indexing."""
        try:
            indexing_num = 0
            node = subscript_node
            # count indexing number
            while hasattr(node, "value"):
                indexing_num += 1
                node = node.value
            # if chain indexing is used in the code, and the indexing number is no less than two, the rule is violated.
            if(
                hasattr(node, "name")
                and node.name in self._subscript_types
                and self._subscript_types[node.name] == "pd.DataFrame"
                and indexing_num >= 2
            ):
                self.add_message("chain-indexing-pandas", node=subscript_node)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, subscript_node)



"""Class which tests DeterministicAlgorithmChecker."""
import astroid
import pylint.testutils
import dslinter


class TestDeterministicAlgorithmChecker(pylint.testutils.CheckerTestCase):
    """Class which tests DeterministicAlgorithmChecker."""

    CHECKER_CLASS = dslinter.plugin.DeterministicAlgorithmChecker

    def test_with_deterministic_option_set(self):
        """Test whether no message is added if the deterministic algorithm option is used."""
        script = """
        import torch #@
        torch.use_deterministic_algorithms(True)
        
        if __name__ == '__main__':
            pass
        """
        module = astroid.parse(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)

    def test_without_deterministic_option_set(self):
        """Test whether a message is added if the deterministic algorithm option is not used"""
        script = """
        import torch #@
        torch.randn(10).index_copy(0, torch.tensor([0]), torch.randn(1))
        
        if __name__ == '__main__':
            pass
        """
        module = astroid.parse(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="deterministic-pytorch", node=module)):
            self.checker.visit_module(module)



"""Checker checks whether scaler is added before scaling-sensitive operations."""
import traceback
from typing import List
import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker
from dslinter.utils.exception_handler import ExceptionHandler
from dslinter.utils.ast import AssignUtil


class ScalerMissingScikitLearnChecker(BaseChecker):
    """Checker checks whether scaler is added before scaling-sensitive operations."""

    __implements__ = IAstroidChecker

    name = "scaler-missing-scikitlearn"
    priority = -1
    msgs = {
        "W5505": (
            "Scaler is not used before scaling-sensitive operation",
            "scaler-missing-scikitlearn",
            "To ensure a good result, use feature scaling before scaling-sensitive operation."
        ),
    }
    options = ()

    PIPELINE = [
        "make_pipeline",
        "Pipeline",
    ]

    SCALING_SENSITIVE_OPERATIONS = [
        "PCA",
        "KernelPCA",
        "SparsePCA",
        "IncrementalPCA",
        "LinearSVC",
        "LinearSVR",
        "NuSVC",
        "NuSVR",
        "OneClassSVM",
        "SVC",
        "SVR",
        "SGDClassifier",
        "SGDOneClassSVM",
        "SGDRegressor",
        "MLPClassifier",
        "MLPRegressor"
    ]

    SCALER = ["RobustScaler", "StandardScaler", "MaxAbsScaler", "MinMaxScaler",]

    LEARNING_FUNCTIONS: List[str] = [
        "fit",
        "fit_transform",
        "transform",
    ]

    Variables = []

    def visit_call(self, node: astroid.Call):
        """
        When a node is visited, add a message if the rule is violated.
        :param node:
        :return:
        """
        try:
            # If there is no scaler before a scaling-sensitive operarion, the rule is violated.
            # If pipeline is used
            if (
                hasattr(node, "func")
                and hasattr(node.func, "name")
                and node.func.name in self.PIPELINE
                and hasattr(node, "args")
            ):
                has_scaling_sensitive_operation = False
                has_scaler = False
                for arg in node.args:
                    if isinstance(arg, astroid.Call):
                        if self._call_initiates_scaler(arg):
                            has_scaler = True
                        if self._call_initiates_scaling_sensitive_operations(arg):
                            has_scaling_sensitive_operation = True
                            break
                if has_scaling_sensitive_operation is True and has_scaler is False:
                    self.add_message("scaler-missing-scikitlearn", node=node)

            # If pipeline is not used and a scaling-sensitive operation is called
            if (
                    hasattr(node, "func")
                    and hasattr(node.func, "attrname")
                    and node.func.attrname in self.LEARNING_FUNCTIONS
                    and hasattr(node.func, "expr")
                    and self._expr_is_scaling_sensitive_operation(node.func.expr)
                    and hasattr(node, "args")
            ):
                has_scaling_sensitive_operation = True
                has_scaler = False
                for arg in node.args:
                    if isinstance(arg, astroid.Name):
                        values = AssignUtil.assignment_values(arg)
                        for value in values:
                            if (
                                isinstance(value, astroid.Call)
                                and hasattr(value, "func")
                                and hasattr(value.func, "expr")
                            ):
                                if self._expr_is_scaler(value.func.expr):
                                    has_scaler = True
                if has_scaling_sensitive_operation is True and has_scaler is False:
                    self.add_message("scaler-missing-scikitlearn", node=node)

        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, node)
            traceback.print_exc()

    def _call_initiates_scaling_sensitive_operations(self, call: astroid.Call) -> bool:
        """
        Evaluate whether a Call node is initiating an pca.

        :param call: Call to evaluate.
        :return: True when an estimator is initiated.
        """
        return (
                hasattr(call, "func")
                and hasattr(call.func, "name")
                and call.func.name in self.SCALING_SENSITIVE_OPERATIONS
        )

    def _call_initiates_scaler(self, call: astroid.Call) -> bool:
        """
        Evaluate whether a Call node is initiating an scaler.

        :param call: Call to evaluate.
        :return: True when an estimator is initiated.
        """
        return (
                hasattr(call, "func")
                and hasattr(call.func, "name")
                and call.func.name in self.SCALER
        )

    def _expr_is_scaling_sensitive_operation(self, expr: astroid.node_classes.NodeNG) -> bool:
        """
        Evaluate whether the expression is an estimator.

        :param expr: Expression to evaluate.
        :return: True when the expression is an estimator.
        """
        # pylint: disable = line-too-long
        if isinstance(expr, astroid.Call) and self._call_initiates_scaling_sensitive_operations(expr):
            return True

        # If expr is a Name, check whether that name is assigned to an estimator.
        if isinstance(expr, astroid.Name):
            values = AssignUtil.assignment_values(expr)
            for value in values:
                if self._expr_is_scaling_sensitive_operation(value):
                    return True
        return False

    def _expr_is_scaler(self, expr: astroid.node_classes.NodeNG) -> bool:
        """
        Evaluate whether the expression is an estimator.

        :param expr: Expression to evaluate.
        :return: True when the expression is an estimator.
        """
        if isinstance(expr, astroid.Call) and self._call_initiates_scaler(expr):
            return True

        # If expr is a Name, check whether that name is assigned to an estimator.
        if isinstance(expr, astroid.Name):
            values = AssignUtil.assignment_values(expr)
            for value in values:
                if self._expr_is_scaler(value):
                    return True
        return False



"""Class which tests the GradientClearPytorchChecker."""
import astroid
import pylint.testutils

import dslinter


class TestGradientClearPytorchChecker(pylint.testutils.CheckerTestCase):
    """Class which tests the GradientClearPytorchChecker."""

    CHECKER_CLASS = dslinter.plugin.GradientClearPytorchChecker

    def test_gradient_not_clear(self):
        """Message should be added if the optimizer.zero_grad() is not used in pytorch code when loss_fn.backward() and optimizer.step() are used."""
        script = """
            for epoch in range(2):  # loop over the dataset multiple times
            
                running_loss = 0.0
                for i, data in enumerate(trainloader, 0): #@
                    # get the inputs; data is a list of [inputs, labels]
                    inputs, labels = data
            
                    # forward + backward + optimize
                    outputs = net(inputs)
                    loss = criterion(outputs, labels)
                    loss.backward()
                    optimizer.step()
            
                    # print statistics
                    running_loss += loss.item()
                    if i % 2000 == 1999:    # print every 2000 mini-batches
                        print(f'[{epoch + 1}, {i + 1:5d}] loss: {running_loss / 2000:.3f}')
                        running_loss = 0.0        
        """
        for_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="gradient-clear-pytorch", node=for_node)):
            self.checker.visit_for(for_node)

    def test_gradient_clear(self):
        """No message should be added if the loss_fn.backward() and optimizer.step() should be used together with optimizer.zero_grad()."""
        script = """
            for epoch in range(2):  # loop over the dataset multiple times
            
                running_loss = 0.0
                for i, data in enumerate(trainloader, 0): #@
                    # get the inputs; data is a list of [inputs, labels]
                    inputs, labels = data
            
                    # zero the parameter gradients
                    optimizer.zero_grad()
            
                    # forward + backward + optimize
                    outputs = net(inputs)
                    loss = criterion(outputs, labels)
                    loss.backward()
                    optimizer.step()
            
                    # print statistics
                    running_loss += loss.item()
                    if i % 2000 == 1999:    # print every 2000 mini-batches
                        print(f'[{epoch + 1}, {i + 1:5d}] loss: {running_loss / 2000:.3f}')
                        running_loss = 0.0        
        """
        for_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_for(for_node)



import astroid


def check_module_with_library(node, library_name: str):
    while not isinstance(node.parent, astroid.Module):
        node = node.parent
    module = node

    if isinstance(module, astroid.Module):
        for node in module.body:
            if isinstance(node, astroid.Import):
                for name, _ in node.names:
                    if name == library_name:
                        return True
    return False



"""Utility module for working with the Abstract Syntax Tree (AST)."""
from typing import List, Optional, Tuple, Union
import astroid


class ASTUtil:
    """Utility class for working with the Abstract Syntax Tree (AST)."""

    # pylint: disable = line-too-long
    @staticmethod
    def search_nodes(node: astroid.node_classes.NodeNG, type_searched: type) -> List[astroid.node_classes.NodeNG]:
        """
        Search recursively for all nodes of a certain type.

        :param node: Node which is visited.
        :param type_searched: Type of node where is searched for.
        """
        found: List = []
        for child in node.get_children():
            found += ASTUtil.search_nodes(child, type_searched)

        if isinstance(node, type_searched):
            found.append(node)
        return found

    @staticmethod
    def get_source_code(node: astroid.Module) -> str:
        """
        Get the source code of a Module node.

        :param node: Node to get source code from.
        :return: Source code.
        """
        if node.file_bytes is not None:
            # First try if 'file_bytes' attribute is present with the source code.
            return node.file_bytes.decode("utf-8")
        if node.file is not None:
            # Otherwise, read the entire file stated in the 'file' attribute.
            with open(node.file, "r", encoding="utf-8") as file:
                return file.read()
        raise Exception("Could not retrieve the source code of the module.")

    @staticmethod
    def search_body_parent(node: astroid.node_classes.NodeNG) -> astroid.node_classes.NodeNG:
        """
        Search the parent of the body block a node is part of.

        :param node: Node to search the parent of the body block of.
        :return: Parent node of the body block.
        """
        if hasattr(node, "body"):
            return node
        return ASTUtil.search_body_parent(node.parent)

    @staticmethod
    def search_module(node: astroid.node_classes.NodeNG) -> astroid.Module:
        """
        Search the module a node is part of.

        :param node: Node to search the module from.
        :return: Module the node is part of.
        """
        if isinstance(node, astroid.Module):
            return node
        return ASTUtil.search_module(node.parent)

    @staticmethod
    def search_body(node: astroid.node_classes.NodeNG) -> List[astroid.node_classes.NodeNG]:
        """
        Search the body block a node is part of.

        :param node: Node to search the body block of.
        :return: Body block the node is part of.
        """
        # noinspection PyUnresolvedReferences
        body = ASTUtil.search_body_parent(node).body
        if isinstance(body, list):
            return body
        else:
            return [body]

    @staticmethod
    def retrieve_keyword_from_list(keywords: List[astroid.Keyword], arg_name: str) -> Optional[astroid.Keyword]:
        """
        Retrieve the keyword with a certain arg from a list of keywords.

        :param keywords: List of keywords.
        :param arg_name: Arg name to look for.
        :return: Keyword node with arg 'arg' or None if it not found.
        """
        for keyword in keywords:
            if keyword.arg == arg_name:
                return keyword
        return None


class AssignUtil:
    """Utility class for working with (Ann)Assign nodes."""

    @staticmethod
    def is_target(name: str, assign: Union[astroid.Assign, astroid.AnnAssign]):
        """
        Evaluate whether 'name' is the target of an (Ann)Assign node.

        :param name: Name to check.
        :param assign: (Ann)Assign node to check.
        :return: True when this (Ann)Assign node has 'name' as a target.
        """
        if isinstance(assign, astroid.Assign):
            for target in assign.targets:
                if hasattr(target, "name") and target.name == name:
                    return True
        if isinstance(assign, astroid.AnnAssign):
            return hasattr(assign.target, "name") and assign.target.name == name
        return False

    @staticmethod
    def assignment_values(name_node: astroid.Name) -> List[astroid.node_classes.NodeNG]:
        """
        Search for the value of the assignment to the name of a Name node.

        :param name_node: Node of which the assignment value is searched.
        :return: Value nodes which are assigned to the name from the Name node.
        """
        name = name_node.name
        function_with_arg, idx = AssignUtil._name_is_arg_from_function(name, name_node)
        if function_with_arg is not None:
            return AssignUtil._function_arg_values(function_with_arg, name, idx)

        return AssignUtil._assign_values_in_body_of_parents(name, name_node)

    @staticmethod
    def _assign_values_in_body_of_parents(
        name: str, node: astroid.node_classes.NodeNG
    ) -> List[astroid.node_classes.NodeNG]:
        """
        Search for the value of the assignment to the name of a Name node in body of parents.

        :param name: Name to look for.
        :param node: Node of which is searched in the body of its parents.
        :return: The value node of the assignment or None when it is not found.
        """
        values = []
        body_block = ASTUtil.search_body(node)
        for child in body_block:
            if isinstance(child, (astroid.AnnAssign, astroid.Assign)) \
                    and AssignUtil.is_target(name, child):
                values.append(child.value)
        if hasattr(node, "parent") and node.parent is not None:
            return AssignUtil._assign_values_in_body_of_parents(name, node.parent)
        return values

    @staticmethod
    def _name_is_arg_from_function(
        name: str, node: astroid.node_classes.NodeNG
    ) -> Tuple[Optional[astroid.FunctionDef], int]:
        """
        Get the FunctionDef the name is part of (if it exists).

        :param name: Name to look for.
        :param node: Node to look for FunctionDef nodes in.
        :return: FunctionDef with arg 'name' or None when not found and the position of the arg.
        """
        if isinstance(node, astroid.FunctionDef):
            for idx, arg in enumerate(node.args.args):
                if hasattr(arg, "name") and arg.name == name:
                    return node, idx
            return None, -1
        if hasattr(node, "parent"):
            return AssignUtil._name_is_arg_from_function(name, node.parent)
        return None, -1

    # noinspection PyUnresolvedReferences
    @staticmethod
    def _function_arg_values(node: astroid.FunctionDef, arg_name: str, arg_position: int):
        """
        Search the values a certain argument of a function gets assigned.

        Calls to the function will be searched for in the parent of the FunctionDef node.

        :param node: FunctionDef node which contains the argument.
        :param arg_name: Name of the argument.
        :param arg_position: Position of the argument.
        :return: All values this argument gets assigned.
        """
        values = []
        for call_node in ASTUtil.search_nodes(node.parent, astroid.Call):
            if hasattr(call_node.func, "name") and call_node.func.name == node.name:
                if call_node.args is not None and len(call_node.args) > arg_position:
                    values.append(call_node.args[arg_position])
                if call_node.keywords is not None:
                    keyword = ASTUtil.retrieve_keyword_from_list(call_node.keywords, arg_name)
                    if keyword is not None:
                        values.append(keyword.value)
        return values

    @staticmethod
    def get_assigned_target_names(node: astroid.node_classes.NodeNG) -> List[str]:
        """
        Get the target names of all assign nodes in the body of a node.

        :param node: Node to get the target names from.
        :return: Target names.
        """
        assigned_names = []
        if not hasattr(node, "body"):
            return []

        for body_node in node.body:
            if isinstance(body_node, astroid.Assign):
                for target in body_node.targets:
                    assigned_names.append(AssignUtil._get_target_name(target))
            if isinstance(body_node, astroid.AnnAssign):
                assigned_names.append(AssignUtil._get_target_name(body_node.target))
            else:
                assigned_names += AssignUtil.get_assigned_target_names(body_node)
        return assigned_names

    @staticmethod
    def _get_target_name(target: astroid.node_classes.NodeNG) -> str:
        """
        Get the name attribute of a node listed as target.

        :param target: Node to get the name from.
        :return: Name.
        """
        if hasattr(target, "name"):
            return target.name
        if hasattr(target, "value"):
            return AssignUtil._get_target_name(target.value)
        raise Exception("Target name cannot be retrieved.")



"""Functions for getting the signatures of Classes."""
# pylint: disable = line-too-long
import inspect
import pickle
from typing import List


def save_hyperparameter(classes: List, path: str):
    """Functions for getting the signatures of Classes."""
    # Collect all signatures of the learning classes.
    signatures = []
    for c in classes:
        try:
            signatures.append((c.__name__, inspect.signature(c)))
        except ValueError as ex:
            print(ex)

    # Construct the dict with hyperparameters from the signatures.
    hyperparameters = {}
    for class_name, signature in signatures:
        keywords_amount = len(signature.parameters)
        keywords = list(signature.parameters.keys())
        hyperparameters[class_name] = {"positional": keywords_amount, "keywords": keywords}

    print(hyperparameters)

    # Write the pickled hyperparameters dict to disk and verify it.
    with open(path, "wb") as file_handler:
        pickle.dump(hyperparameters, file_handler)
        print("The pickle with all hyperparameters is written to disk.")
    with open(path, "rb") as file_handler:
        hyperparameters_loaded = pickle.load(file_handler)
        assert hyperparameters == hyperparameters_loaded

    print("Done!")



"""Checker which checks whether the parameters for merge operations are set."""
import astroid as astroid
from pylint.interfaces import IAstroidChecker
from pylint.checkers import BaseChecker
from typing import Dict

from dslinter.utils.exception_handler import ExceptionHandler
from dslinter.utils.type_inference import TypeInference


class MergeParameterPandasChecker(BaseChecker):
    """Checker which checks whether the parameters for merge operations are set."""

    __implements__ = IAstroidChecker

    name = "merge-parameter-pandas"
    priority = -1
    msgs = {
        "R5505": (
            "Parameters for merge operations are not set.",
            "merge-parameter-pandas",
            "Parameters 'how', 'on' and 'validate' should be set for merge operations to ensure the correct usage of merging."
        )
    }
    options = ()

    # [variable name, inferred type of object the function is called on]
    _subscript_types: Dict[str, str] = {}

    def visit_module(self, module: astroid.Module):
        """Visit module and infer which library the variables are from. """
        try:
            self._subscript_types = TypeInference.infer_library_variable_most_recent_types(module)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, module)

    def visit_call(self, call_node: astroid.Call):
        """Visit call node and check whether the parameters are set."""
        # call on pandas dataframe object && name "merge" && check parameter
        try:
            if(
                hasattr(call_node.func, "attrname")
                and call_node.func.attrname == "merge"
                and hasattr(call_node.func, "expr")
                and hasattr(call_node.func.expr, "name")
                and call_node.func.expr.name in self._subscript_types
                and self._subscript_types[call_node.func.expr.name] in ["pd.DataFrame", "pandas.DataFrame"]
            ):
                kws = [kw.arg for kw in call_node.keywords if hasattr(kw, "arg")]
                if(
                    "how" in kws
                    and "on" in kws
                    and "validate" in kws
                ):
                    pass
                else:
                    self.add_message("merge-parameter-pandas", node=call_node)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, call_node)



"""Hyperparameter checker checks whether all hyperparameters for learning algorithms are set."""
from typing import List, Dict
import astroid
from pylint.lint import PyLinter
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker
from dslinter.utils.exception_handler import ExceptionHandler
from dslinter.utils.resources import Resources


class HyperparameterChecker(BaseChecker):
    """Hyperparameter checker checks whether all hyperparameters for learning algorithms are set."""

    __implements__ = IAstroidChecker

    def __init__(self, linter: PyLinter = None) -> None:
        super().__init__(linter)
        self.HYPERPARAMETERS_MAIN = {}
        self.HYPERPARAMETER_RESOURCE = ""
        self.MESSAGE = ""
        self.LIBRARY = ""
        self.call_types = {}

    def visit_importfrom(self, node: astroid.ImportFrom):
        try:
            for name, _ in node.names:
                self.call_types[name] = node.modname.split('.')[0]
        except:
            ExceptionHandler.handle(self, node)

    def visit_call(self, node: astroid.Call):
        """
        When a Call node is visited, check whether hyperparameters are set.

        In strict mode, all hyperparameters should be set.
        In non-strict mode, function calls to learning functions should either contain all
        hyperparameters defined in HYPERPARAMETERS_MAIN or have at least one hyperparameter defined.

        :param node: Node which is visited.
        """
        try:
            if(
                hasattr(node, "func")
                and hasattr(node.func, "name")
            ):
                self.hyperparameter_in_class(node, node.func.name)
        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, node)

    def hyperparameter_in_class(self, node: astroid.Call, function_name: str):
        """Cheches whether the required hyperparameters are used in the class."""

        hyperparams_all = Resources.get_hyperparameters(self.HYPERPARAMETER_RESOURCE)

        strict_hyperparameters = ""
        if self.LIBRARY == "scikitlearn": # strict mode
            strict_hyperparameters = self.config.strict_hyperparameters_scikitlearn
        elif self.LIBRARY == "tensorflow":
            strict_hyperparameters = self.config.strict_hyperparameters_tensorflow
            if function_name in self.call_types and self.call_types[function_name] not in ["tf", "tensorflow"]:
                return
        elif self.LIBRARY == "pytorch":
            strict_hyperparameters = self.config.strict_hyperparameters_pytorch
            if function_name in self.call_types and self.call_types[function_name] != "torch":
                return

        if function_name in hyperparams_all:  # pylint: disable=unsupported-membership-test
            if strict_hyperparameters:
                if not self.has_required_hyperparameters(node, hyperparams_all, function_name):
                    self.add_message(self.MESSAGE, node=node)
            else:  # non-strict mode
                if (
                    function_name in self.HYPERPARAMETERS_MAIN
                    # pylint: disable = line-too-long
                    and not self.has_required_hyperparameters(node, self.HYPERPARAMETERS_MAIN, function_name)
                ):
                    self.add_message(self.MESSAGE, node=node)
                elif len(node.args) == 0 and node.keywords is None:
                    self.add_message(self.MESSAGE, node=node)

    def has_required_hyperparameters(self, node: astroid.Call, hyperparameters: Dict, name: str):
        """
        Evaluate whether a function call has all required hyperparameters defined.

        :param node: Node which is visited.
        :param hyperparameters: Dict of functions with their required hyperparameters.
        :return: True when all required hyperparameters are defined.
        """
        return len(node.args) >= hyperparameters[name]["positional"] or self.has_keywords(
            node.keywords, hyperparameters[name]["keywords"]
        )

    @staticmethod
    def has_keywords(keywords: List[astroid.Keyword], keywords_goal: List[str]) -> bool:
        """
        Check if a list of keywords contains certain keywords.

        :param keywords: List of keywords.
        :param keywords_goal: Name of the keywords which are checked against.
        :return: True if keywords are present, False if they are not.
        """
        if keywords is None:
            return False

        found = 0
        for keyword in keywords:
            if keyword.arg in keywords_goal:
                found += 1

        return found == len(keywords_goal)



"""Class which tests RandomnessControllingTensorflowChecker"""
import astroid
import pylint.testutils
import dslinter


class TestRandomnessControlTensorflowChecker(pylint.testutils.CheckerTestCase):
    """Class which tests RandomnessControllingTensorflowChecker"""

    CHECKER_CLASS = dslinter.plugin.RandomnessControlTensorflowChecker

    def test_with_tensorflow_randomness_control(self):
        """Tests whether no message is added if manual seed is set."""
        script = """
        import tensorflow as tf #@
        tf.random.set_seed(0)
        tf.random.uniform([1])
                
        if __name__ == '__main__':
            pass
        """
        module = astroid.parse(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)

    def test_without_tensorflow_randomness_control(self):
        """Tests whether a message is added if manual seed is not set"""
        script = """
        import tensorflow as tf #@
        tf.random.uniform([1])
                
        if __name__ == '__main__':
            pass
        """
        module = astroid.parse(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="randomness-control-tensorflow", node=module)):
            self.checker.visit_module(module)



"""Class which tests TestMergeParameterPandasChecker."""
import astroid
import pylint.testutils

import dslinter.plugin


class TestMergeParameterPandasChecker(pylint.testutils.CheckerTestCase):
    """Class which tests TestMergeParameterPandasChecker."""

    CHECKER_CLASS = dslinter.plugin.MergeParameterPandasChecker

    def test_merge_parameter_pandas_checker_not_set(self):
        """Message should be added if the parameters for merge operations are not set."""
        script = """
            import pandas as pd
            df1 = pd.DataFrame({'key': ['foo', 'bar', 'baz', 'foo'],
                                'value': [1, 2, 3, 5]})
            df2 = pd.DataFrame({'key': ['foo', 'bar', 'baz', 'foo'],
                                'value': [5, 6, 7, 8]})    
            df3 = df1.merge(df2)   #@                   
        """
        module = astroid.parse(script)
        assign_node = astroid.extract_node(script)
        call_node = assign_node.value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="merge-parameter-pandas", node=call_node)):
            self.checker.visit_module(module)
            self.checker.visit_call(call_node)

    def test_merge_parameter_pandas_checker_set(self):
        """No message should be added if the parameters for merge operations are set."""
        script = """
            import pandas as pd
            df1 = pd.DataFrame({'key': ['foo', 'bar', 'baz', 'foo'],
                                'value': [1, 2, 3, 5]})
            df2 = pd.DataFrame({'key': ['foo', 'bar', 'baz', 'foo'],
                                'value': [5, 6, 7, 8]})                  
            df3 = df1.merge(df2, how='inner', on='key', validate='m:m')     #@
        """
        module = astroid.parse(script)
        assign_node = astroid.extract_node(script)
        call_node = assign_node.value
        with self.assertNoMessages():
            self.checker.visit_module(module)
            self.checker.visit_call(call_node)



"""Check whether there is unnecessary iteration in Pandas code."""
from typing import List
from typing import Dict
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker
import astroid
from dslinter.utils.exception_handler import ExceptionHandler
from dslinter.utils.ast import AssignUtil
from dslinter.utils.type_inference import TypeInference


class UnnecessaryIterationPandasChecker(BaseChecker):
    """Check whether there is unnecessary iteration in Pandas code."""
    __implements__ = IAstroidChecker

    name = "unnecessary-iteration-pandas"
    priority = -1
    msgs = {
        "R5501": (
            "Iterating through a DataFrame.",
            "unnecessary-iteration-pandas",
            "Iteration through a DataFrame is generally slow and should be avoided.",
        ),
        "W5501": (
            "Iterated object is modified.",
            "dataframe-iteration-modification-pandas",
            "An object where is iterated over should not be modified.",
        ),
    }
    options = ()

    # [node, inferred type of object the function is called on]
    _call_types: Dict[astroid.Call, str] = {}

    def visit_module(self, node: astroid.Module):
        """
        When an Module node is visited, scan for Call nodes and get type the function is called on.

        :param node: Node which is visited.
        """
        try:
            # noinspection PyTypeChecker
            self._call_types = TypeInference.infer_types(node,
                                                         astroid.Call,
                                                         lambda x: x.func.expr.name)
        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, node)

    def visit_call(self, node: astroid.Call):
        """
        When a Call node is visited, add messages if it violated the defined rules.

        :param node: Node which is visited.
        """
        try:
            if self._iterating_through_dataframe(node):
                self.add_message("dataframe-iteration-modification-pandas", node=node)
        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, node)

    def _iterating_through_dataframe(self, node: astroid.Call) -> bool:
        """
        Evaluate whether there is iterated through a DataFrame.

        :param node: Node which is visited.
        :return: True when there is iterated through a DataFrame.
        """
        return (
            isinstance(node.parent, astroid.For)
            and node not in node.parent.body
            and node in self._call_types
            and (
                self._call_types[node] == '"pandas.core.frame.DataFrame"'
                or self._call_types[node] == '"pyspark.sql.dataframe.DataFrame"'
            )
        )

    def visit_for(self, node: astroid.For):
        """
        When a For node is visited, check for dataframe-iteration-modification violations.

        :param node: Node which is visited.
        """
        try:
            if not (
                isinstance(node.iter, astroid.Call)
                and node.iter in self._call_types
                and (
                    self._call_types[node.iter] == '"pandas.core.frame.DataFrame"'
                    or self._call_types[node.iter] == '"pyspark.sql.dataframe.DataFrame"'
                )
            ):
                return

            for_targets = self._get_for_targets(node)
            assigned = AssignUtil.get_assigned_target_names(node)
            modified_iterated_targets = any(target in for_targets for target in assigned)

            if modified_iterated_targets:
                self.add_message("dataframe-iteration-modification-pandas", node=node)
        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, node)

    @staticmethod
    def _get_for_targets(node: astroid.For) -> List[str]:
        """
        Get the target names of the for-loop definition.

        :param node: For node to get the target names from.
        :return: Target names.
        """
        target_names = []
        if isinstance(node.target, astroid.Tuple):
            for element in node.target.elts:
                if isinstance(element, astroid.AssignName):
                    target_names.append(element.name)
        elif isinstance(node.target, astroid.AssignName):
            target_names.append(node.target.name)
        return target_names



"""Checker which checks whether training mode and evaluation mode is toggling properly in pytorch code."""
import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker

from dslinter.utils.exception_handler import ExceptionHandler


class ModeTogglingPytorchChecker(BaseChecker):
    """Checker which checks whether training mode and evaluation mode is toggling properly in pytorch code."""

    __implements__ = IAstroidChecker

    name = "mode-toggling-pytorch"
    priority = -1
    msgs = {
        "W9998": (
            "The training mode did not toggle back in time in the pytorch code.",
            "mode-toggling-pytorch",
            "Developers should call the training mode in the right place to avoid forgetting to switch back to the training mode after the inference step."
        )
    }
    options = ()

    def visit_for(self, for_node: astroid.For):
        """
        When a For node is visited, check whether it violated the rule in this checker.
        :param for_node: The node which is visited.
        """
        try:
            _has_train = False
            _has_eval = False
            for node in for_node.body:
                if isinstance(node, astroid.Expr):
                    if isinstance(node.value, astroid.Call) and hasattr(node.value.func, "attrname"):
                        if node.value.func.attrname == "train":
                            _has_train = True
                        if node.value.func.attrname == "eval":
                            _has_eval = True
                if isinstance(node, astroid.If):
                    for if_node in node.body:
                        if(
                            isinstance(if_node, astroid.Expr)
                            and isinstance(if_node.value, astroid.Call)
                            and hasattr(if_node.value.func, "attrname")
                        ):
                            if if_node.value.func.attrname == "train":
                                _has_train = True
                            if if_node.value.func.attrname == "eval":
                                _has_eval = True

            if _has_eval is True and _has_train is False:
                self.add_message("mode-toggling-pytorch", node=for_node)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, for_node)



"""Class which tests DependentThresholdPytorchChecker."""
import astroid
import pylint.testutils
import dslinter.plugin


class TestDependentThresholdPytorchChecker(pylint.testutils.CheckerTestCase):
    """Class which tests DependentThresholdPytorchChecker."""

    CHECKER_CLASS = dslinter.plugin.DependentThresholdPytorchChecker

    def test_with_auc_used(self):
        script = """
        from torchmetrics import F1Score
        target = torch.tensor([0, 1, 2, 0, 1, 2])
        preds = torch.tensor([0, 2, 1, 0, 0, 1])
        f1 = F1Score(num_classes=3)
        f1(preds, target)

        from torchmetrics import AUROC
        preds = torch.tensor([0.13, 0.26, 0.08, 0.19, 0.34])
        target = torch.tensor([0, 0, 1, 1, 1])
        auroc = AUROC(pos_label=1)
        auroc(preds, target)
        """
        module = astroid.parse(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)

    def test_with_only_f1_score_used(self):
        script = """
        from torchmetrics import F1Score
        target = torch.tensor([0, 1, 2, 0, 1, 2])
        preds = torch.tensor([0, 2, 1, 0, 0, 1])
        f1 = F1Score(num_classes=3)
        f1(preds, target)
        """
        module = astroid.parse(script)
        f1_score_node = module.body[-2].value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="dependent-threshold-pytorch", node=f1_score_node)):
            self.checker.visit_module(module)



"""Checker which checks rules for preventing data leakage between training and test data."""
from typing import List

import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker

from dslinter.utils.ast import AssignUtil
from dslinter.utils.exception_handler import ExceptionHandler
from dslinter.utils.resources import Resources


class DataLeakageScikitLearnChecker(BaseChecker):
    """Checker which checks rules for preventing data leakage between training and test data."""

    __implements__ = IAstroidChecker

    name = "data-leakage-scikitlearn"
    priority = -1
    msgs = {
        "W5518": (
            "There are both preprocessing and estimation operations in the code, but they are not used in a pipeline.",
            "data-leakage-scikitlearn",
            "Scikit-learn preprocessors and estimators should be used inside pipelines, to prevent data leakage between training and test data.",
        ),
    }
    options = ()

    LEARNING_FUNCTIONS: List[str] = [
        "fit",
        "fit_predict",
        "fit_transform",
        "predict",
        "score",
        "transform",
    ]

    PREPROCESSING_CLASSES = [
        "FunctionTransformer",
        "KBinsDiscretizer",
        "KernelCenterer",
        "LabelBinarizer",
        "LabelEncoder",
        "MultiLabelBinarizer",
        "MaxAbsScaler",
        "MinMaxScaler",
        "OneHotEncoder",
        "OrdinalEncoder",
        "PolynomialFeatures",
        "PowerTransformer",
        "QuantileTransformer",
        "RobustScaler",
        "StandardScaler",
        "SelectKBest",
    ]

    def visit_call(self, call_node: astroid.Call):
        """
        When a Call node is visited, check whether it violated the rules in this checker.

        :param call_node: The node which is visited.
        """
        try:
            # If the learning function is called on an estimator, rule is violated.
            if (
                call_node.func is not None
                and hasattr(call_node.func, "attrname")
                and call_node.func.attrname in self.LEARNING_FUNCTIONS
                and hasattr(call_node.func, "expr")
                and self._expr_is_estimator(call_node.func.expr)
                and hasattr(call_node, "args")
            ):
                has_learning_function = True
                has_preprocessing_function = False
                for arg in call_node.args:
                    if isinstance(arg, astroid.Name):
                        values = AssignUtil.assignment_values(arg)
                        for value in values:
                            if (
                                isinstance(value, astroid.Call)
                                and hasattr(value, "func")
                                and hasattr(value.func, "expr")
                            ):
                                if self._expr_is_preprocessor(value.func.expr):
                                    has_preprocessing_function = True
                if has_learning_function is True and has_preprocessing_function is True:
                    self.add_message("data-leakage-scikitlearn", node=call_node)

        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, call_node)

    @staticmethod
    def _expr_is_estimator(expr: astroid.node_classes.NodeNG) -> bool:
        """
        Evaluate whether the expression is an estimator.

        :param expr: Expression to evaluate.
        :return: True when the expression is an estimator.
        """
        if isinstance(expr, astroid.Call) \
                and DataLeakageScikitLearnChecker._call_initiates_estimator(expr):
            return True

        # If expr is a Name, check whether that name is assigned to an estimator.
        if isinstance(expr, astroid.Name):
            values = AssignUtil.assignment_values(expr)
            for value in values:
                if DataLeakageScikitLearnChecker._expr_is_estimator(value):
                    return True
        return False

    @staticmethod
    def _call_initiates_estimator(call: astroid.Call) -> bool:
        """
        Evaluate whether a Call node is initiating an estimator.

        :param call: Call to evaluate.
        :return: True when an estimator is initiated.
        """
        return (
                call.func is not None
                and hasattr(call.func, "name")
                and call.func.name in DataLeakageScikitLearnChecker._get_estimator_classes()
        )

    @staticmethod
    def _get_estimator_classes() -> List[str]:
        """
        Get all estimator classes.

        The list contains all learning classes which do something in the
        fit function from sklearn.

        :return: List of estimator classes.
        """
        # pylint: disable = line-too-long
        learning_classes = list(Resources.get_hyperparameters("hyperparameters_scikitlearn_dict.pickle").keys())
        estimator_classes = learning_classes
        assert None not in estimator_classes
        return estimator_classes

    def _call_initiates_preprocessor(self, call: astroid.Call) -> bool:
        """
        Evaluate whether a Call node is initiating an scaler.

        :param call: Call to evaluate.
        :return: True when an estimator is initiated.
        """
        return (
                hasattr(call, "func")
                and hasattr(call.func, "name")
                and call.func.name in self.PREPROCESSING_CLASSES
        )

    def _expr_is_preprocessor(self, expr: astroid.node_classes.NodeNG) -> bool:
        """
        Evaluate whether the expression is an estimator.

        :param expr: Expression to evaluate.
        :return: True when the expression is an estimator.
        """
        if isinstance(expr, astroid.Call) and self._call_initiates_preprocessor(expr):
            return True

        # If expr is a Name, check whether that name is assigned to an estimator.
        if isinstance(expr, astroid.Name):
            values = AssignUtil.assignment_values(expr)
            for value in values:
                if self._expr_is_preprocessor(value):
                    return True
        return False



"""Class which tests RandomnessControllingNumpyChecker"""
import astroid
import pylint.testutils
import dslinter


class TestRandomnessControlNumpyChecker(pylint.testutils.CheckerTestCase):
    """Class which tests RandomnessControllingNumpyChecker"""

    CHECKER_CLASS = dslinter.plugin.RandomnessControlNumpyChecker

    def test_ml_code_with_numpy_randomness_control(self):
        """Tests whether no message is added if manual seed is set."""
        script = """
        import numpy as np #@
        import torch #@
        np.random.seed(0)
        np.random.rand(4)
                
        if __name__ == '__main__':
            pass
        """
        module = astroid.parse(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)

    def test_non_ml_code_with_numpy_randomness_control(self):
        """Tests whether no message is added if manual seed is set."""
        script = """
        import numpy as np #@
        np.random.seed(0)
        np.random.rand(4)
                
        if __name__ == '__main__':
            pass
        """
        module = astroid.parse(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)

    def test_ml_code_without_numpy_randomness_control(self):
        """Tests whether a message is added if manual seed is not set"""
        script = """
        import numpy as np #@
        import torch #@
        np.random.rand(4) 
                
        if __name__ == '__main__':
            pass
        """
        module = astroid.parse(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="randomness-control-numpy", node=module)):
            self.checker.visit_module(module)

    def test_non_ml_code_without_numpy_randomness_control(self):
        """Tests whether a message is added if manual seed is not set"""
        script = """
        import numpy as np #@
        np.random.rand(4) 
                
        if __name__ == '__main__':
            pass
        """
        module = astroid.parse(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)



"""Plugin for the pylint package which allows static code analysis on data science code."""
from __future__ import absolute_import

import sys

from dslinter import plugin

if sys.version_info < (3,):
    raise DeprecationWarning("Python 2 is not supported. Please migrate to Python 3!")

register = plugin.register  # pylint: disable=invalid-name



"""Import checker which checks whether imports are bound to local names by the conventions."""
import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker

from dslinter.utils.exception_handler import ExceptionHandler


class ImportChecker(BaseChecker):
    """Import checker which checks whether imports are bound to local names by the conventions."""

    __implements__ = IAstroidChecker

    name = "import"
    priority = -1
    msgs = {
        "C5501": (
            "The import of pandas doesn't bound to 'pd'.",
            "import-pandas",
            "The pandas module should be imported as 'pd'.",
        ),
        "C5502": (
            "The import of numpy doesn't bound to 'np'.",
            "import-numpy",
            "The numpy module should be imported as 'np'.",
        ),
        "C5503": (
            "The Import of matplotlib.pyplot doesn't bound to 'plt'.",
            "import-pyplot",
            "The matplotlib.pyplot module should be imported as 'plt'.",
        ),
        "C5504": (
            "The import from sklearn module has an alias.",
            "import-sklearn",
            "Imports from sklearn modules should not have an alias.",
        ),
        "C5505": (
            "The import of tensorflow doesn't bound to 'tf'.",
            "import-tensorflow",
            "The tensorflow module should be imported as 'tf' ",
        ),
        "C5506": (
            "The import of pytorch has an alias.",
            "import-pytorch",
            "The pytorch module should not have an alias",
        )
    }
    options = ()

    def visit_import(self, node: astroid.Import):
        """
        When an Import node is visited, check if it follows the conventions.

        :param node: Node which is visited.
        """
        try:
            for name, alias in node.names:
                if name == "pandas" and alias != "pd":
                    self.add_message("import-pandas", node=node)
                elif name == "numpy" and alias != "np":
                    self.add_message("import-numpy", node=node)
                elif name == "matplotlib.pyplot" and alias != "plt":
                    self.add_message("import-pyplot", node=node)
                elif name == "tensorflow" and alias != "tf":
                    self.add_message("import-tensorflow", node=node)
                elif name == "torch" and alias is not None:
                    self.add_message("import-pytorch", node=node)
        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, node)

    def visit_importfrom(self, node: astroid.ImportFrom):
        """
        When an ImportFrom node is visited, check if it follows the conventions.

        :param node: Node which is visited.
        """
        try:
            if node.modname[:7] == "sklearn":
                for _, alias in node.names:
                    if alias is not None:
                        self.add_message("import-sklearn", node=node)
        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, node)



"""Class which tests ColumnSelectionPandasChecker."""
import pylint.testutils
import astroid
import dslinter.plugin


class TestColumnSelectionPandasChecker(pylint.testutils.CheckerTestCase):
    """Class which tests ColumnSelectionPandasChecker."""

    CHECKER_CLASS = dslinter.plugin.ColumnSelectionPandasChecker

    def test_column_not_selected(self):
        """A message should be added if there is no column selection after the dataframe is imported."""
        script = """
        import pandas as pd
        df = pd.read_csv('data.csv')
        """
        module = astroid.parse(script)
        assign_node = module.body[-1]
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="column-selection-pandas", node=assign_node)):
            self.checker.visit_module(module)

    def test_column_selected1(self):
        """No message should be added if there is column selection after the dataframe is imported."""
        script = """
        import pandas as pd
        df = pd.read_csv('data.csv')
        df = df[['col1', 'col2', 'col3']]
        """
        module = astroid.parse(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)

    def test_column_selected2(self):
        """No message should be added if there is column selection after the dataframe is imported."""
        script = """
        import pandas as pd
        df = pd.read_csv('data.csv')
        print(df)
        df = df[['col1', 'col2', 'col3']]
        """
        module = astroid.parse(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)

    def test_a_column_not_selected(self):
        """A message should be added if some imported dataframe are not selected by columns."""
        script = """
        import pandas as pd
        df1 = pd.read_csv('data1.csv')
        df2 = pd.read_csv('data2.csv')
        df1 = df1[['col1', 'col2', 'col3']]
        """
        module = astroid.parse(script)
        assign_node = module.body[-2]
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="column-selection-pandas", node=assign_node)):
            self.checker.visit_module(module)

    def test_all_columns_selected(self):
        """No message should be added if every imported dataframe is selected by columns."""
        script = """
        import pandas as pd
        df1 = pd.read_csv('data1.csv')
        df2 = pd.read_csv('data2.csv')
        df1 = df1[['col1', 'col2', 'col3']]
        df3 = df2[['col1', 'col2']]
        """
        module = astroid.parse(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)

    def test_violation_in_function_def(self):
        script = """
        import os
        from unittest import TestCase
        import pandas as pd
        from typing import List
        
        from lexnlp.extract.common.base_path import lexnlp_test_path
        from lexnlp.extract.common.annotations.law_annotation import LawAnnotation
        # pylint:disable=no-name-in-module
        from lexnlp.extract.de.laws import LawsParser, get_laws
        from lexnlp.tests.typed_annotations_tests import TypedAnnotationsTester
        
        class TestParseDeLaws(TestCase):
            def setup_parser(): #@
                base_path = os.path.join(lexnlp_test_path, 'lexnlp/extract/de/laws/')
                gesetze_df = pd.read_csv(os.path.join(os.path.dirname(__file__),
                                                      base_path + 'gesetze_list.csv'),
                                         encoding="utf-8")
            
                # verordnungen_df = pd.read_csv(os.path.join(os.path.dirname(__file__),
                #                                           base_path + 'verordnungen_list.csv'),
                #                              encoding="utf-8")
            
                # concept_df = pd.read_csv(os.path.join(os.path.dirname(__file__),
                #                                      base_path + 'de_concept_sample.csv'),
                #                         encoding="utf-8")
            
                law_parser = LawsParser(gesetze_df,
                                        verordnungen_df,
                                        concept_df)
                return law_parser
        
        parser = setup_parser()
        """
        module = astroid.parse(script)
        functiondef_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="column-selection-pandas", node=functiondef_node.body[1])):
            self.checker.visit_module(module)
            self.checker.visit_functiondef(functiondef_node)



"""Test checks whether the memory is released in time in Tensorflow code."""
import astroid
import pylint.testutils
import dslinter.plugin


class TestMemoryReleaseTensorflowChecker(pylint.testutils.CheckerTestCase):
    """Test checks whether the memory is released in time in Tensorflow code."""

    CHECKER_CLASS = dslinter.plugin.MemoryReleaseTensorflowChecker

    def test_iteration(self):
        """Check whether a message is added is there is no clear_session() before model creation in the loop"""
        script = """
            import tensorflow as tf
            for _ in range(100): #@
                model = tf.keras.Sequential([tf.keras.layers.Dense(10) for _ in range(10)]) 
        """
        module_tree = astroid.parse(script)
        for_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="memory-release-tensorflow", node=for_node),):
            self.checker.visit_module(module_tree)
            self.checker.visit_for(for_node)

    def test_memory_clear_in_iteration(self):
        """Check whether no message is added is there is a clear_session() before model creation in the loop"""
        script = """
            import tensorflow as tf
            for _ in range(100): #@
                tf.keras.backend.clear_session()
                model = tf.keras.Sequential([tf.keras.layers.Dense(10) for _ in range(10)])
        """
        module_tree = astroid.parse(script)
        for_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_module(module_tree)
            self.checker.visit_for(for_node)



"""DependentThresholdScikitLearnChecker checks whether threshold-independent evaluation
methods(e.g. auc) is used when a threshold-dependent method is used in Scikit-Learn,
because threshold-independent method is always preferred over threshold-dependent method
 in evaluation."""
import astroid
from pylint.interfaces import IAstroidChecker
from pylint.checkers import BaseChecker

from dslinter.utils.exception_handler import ExceptionHandler


class DependentThresholdScikitLearnChecker(BaseChecker):
    """DependentThresholdScikitLearnChecker checks whether threshold-independent evaluation
    methods(e.g. auc) is used when a threshold-dependent method is used in Scikit-Learn,
    because threshold-independent method is always preferred over threshold-dependent method
     in evaluation."""
    __implements__ = IAstroidChecker

    name = "dependent-threshold-scikitlearn"
    priority = -1
    msgs = {
        "W5519": (
            "The F1 Score is used but AUC is not used in the Scikit-learn code.",
            "dependent-threshold-scikitlearn",
            "The threshold independent evaluation method(e.g., AUC) is preferred over threshold dependent method(e.g., F1 Score)."
        )
    }

    options = ()

    def visit_module(self, module):
        """
        When a module node is visited, check whether there is f1 score function called.
        If true, check whether there is auc function.
        If f1 score is called but auc is not called, the rule is violated.
        :param module:
        :return:
        """
        try:
            _has_auc = False
            _has_f1_score = False
            _f1_score_node = None

            for nod in module.body:
                if hasattr(nod, "value") and isinstance(nod.value, astroid.Call):
                    call_node = nod.value
                    if(
                        hasattr(call_node, "func")
                        and hasattr(call_node.func, "name")
                        and call_node.func.name in ["auc", "roc_auc_score"]
                    ):
                        _has_auc = True
                    if(
                        hasattr(call_node, "func")
                        and hasattr(call_node.func, "name")
                        and call_node.func.name == "f1_score"
                    ):
                        _has_f1_score = True
                        _f1_score_node = call_node

            # if f1 score is used but auc is not used
            if _has_f1_score is True and _has_auc is False :
                self.add_message("dependent-threshold-scikitlearn", node=_f1_score_node)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, module)



"""Class which tests the NanChecker."""
import astroid
import pylint.testutils
import dslinter


class TestNanNumpyChecker(pylint.testutils.CheckerTestCase):
    """Class which tests the NanChecker."""

    CHECKER_CLASS = dslinter.plugin.NanNumpyChecker

    def test_nan_left(self):
        """Test whether a message is added when np.nan is compared with something."""
        script = """
        np.nan == b #@
        """
        compare_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="nan-numpy", node=compare_node),):
            self.checker.visit_compare(compare_node)

    def test_nan_right(self):
        """Test whether a message is added when something is compared with np.nan."""
        script = """
        a == np.nan #@
        """
        compare_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="nan-numpy", node=compare_node),):
            self.checker.visit_compare(compare_node)

    def test_nan_both(self):
        """Test whether only one message is added when np.nan is compared with np.nan."""
        script = """
        np.nan == np.nan #@
        """
        compare_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="nan-numpy", node=compare_node),):
            self.checker.visit_compare(compare_node)

    def test_nan_none(self):
        """Test whether no message is added when np.nan is not compared with."""
        script = """
        row == nan #@
        """
        compare_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_compare(compare_node)



"""Class which tests the DataLeakageChecker."""
import astroid
import pylint.testutils
import dslinter


class TestDataLeakageScikitLearnChecker(pylint.testutils.CheckerTestCase):
    """Class which tests the DataLeakageChecker."""

    CHECKER_CLASS = dslinter.plugin.DataLeakageScikitLearnChecker

    def test_pipeline_violation_on_call(self):
        """Message should be added when learning function is called directly on a learning class."""
        script = """
        from sklearn.datasets import load_wine
        from sklearn.model_selection import train_test_split
        from sklearn.metrics import accuracy_score
        from sklearn.svm import SVC
        from sklearn.preprocessing import StandardScaler
        
        # Make a train/tests split using 30% tests size
        RANDOM_STATE = 42
        features, target = load_wine(return_X_y=True)
        X_train, X_test, y_train, y_test = train_test_split(
            features, target, test_size=0.30, random_state=RANDOM_STATE
        )
        scaler = StandardScaler()
        X_train = scaler.fit_transform(X_train)
        X_test = scaler.fit_transform(X_test)
        
        # Fit to data and predict using pipelined GNB and PCA
        SVC().fit(X_train, y_train) #@
        """
        call_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="data-leakage-scikitlearn", node=call_node),):
            self.checker.visit_call(call_node)

    def test_learning_function_without_preprocessor(self):
        """No message should be added when a learning function is called on a non-estimator."""
        script = """
        KMeans().fit(X_train) #@
        """
        call_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    def test_learning_function_on_not_an_estimator(self):
        """No message should be added when a learning function is called on a non-estimator."""
        script = """
        scaler = StandardScaler()
        X_train = scaler.fit_transform(X_train)
        A().fit(X_train) #@
        """
        call_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    def test_pipeline_violation_outside_block(self):
        """Test whether sk-pipeline violation is found when assignment is done outside block."""
        script = """
        scaler = StandardScaler()
        X_train = scaler.fit_transform(X_train)        
        model = KMeans()
        if True:
            model.fit(X_train) #@
        """
        call_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="data-leakage-scikitlearn", node=call_node),):
            self.checker.visit_call(call_node)

    def test_pipeline_violation_on_name(self):
        """Message should be added when learning function is called directly on a learning class."""
        script = """
        scaler = StandardScaler()
        X_train = scaler.fit_transform(X_train)          
        kmeans = KMeans()
        kmeans.fit(X_train) #@
        """
        call_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="data-leakage-scikitlearn", node=call_node),):
            self.checker.visit_call(call_node)

    def test_pipeline_violation_on_name_twice(self):
        """Test calling an estimator by multiple assignments."""
        script = """
        scaler = StandardScaler()
        X_train = scaler.fit_transform(X_train)          
        kmeans = KMeans()
        kmeans2 = kmeans
        kmeans2.fit(X_train) #@
        """
        call_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="data-leakage-scikitlearn", node=call_node),):
            self.checker.visit_call(call_node)

    def test_pipeline_violation_in_function(self):
        """Test whether sk-pipeline violation is found when assignment is a function argument."""
        script = """
        scaler = StandardScaler()
        X_train = scaler.fit_transform(X_train)          
        def f(model):
            model.fit(X_train) #@
        f(KMeans())
        """
        call_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="data-leakage-scikitlearn", node=call_node),):
            self.checker.visit_call(call_node)

    def test_pipeline_violation_in_function_arg_assigned(self):
        """Test calling an estimator within a function, where the argument is assigned."""
        script = """
        scaler = StandardScaler()
        X_train = scaler.fit_transform(X_train)             
        def f(model):
            model.fit(X_train) #@
        kmeans_model = KMeans()
        f(kmeans_model)
        """
        call_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="data-leakage-scikitlearn", node=call_node),):
            self.checker.visit_call(call_node)

    def test_pipeline_violation_in_second_function_argument(self):
        """Test calling an estimator with a second function argument."""
        script = """
        scaler = StandardScaler()
        X_train = scaler.fit_transform(X_train)        
        def f(x, model):
            model.fit(X_train) #@
        f(0, KMeans())
        """
        call_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="data-leakage-scikitlearn", node=call_node),):
            self.checker.visit_call(call_node)



"""Hyperparameter checker checks whether all hyperparameters for learning algorithms are set."""
# pylint: disable = line-too-long
from pylint.interfaces import IAstroidChecker
from pylint.lint import PyLinter
from dslinter.checkers.hyperparameters import HyperparameterChecker


class HyperparameterScikitLearnChecker(HyperparameterChecker):
    """Hyperparameter checker checks whether all hyperparameters for learning algorithms are set."""

    __implements__ = IAstroidChecker

    name = "hyperparameters-scikitlearn"
    priority = -1
    msgs = {
        "R5506": (
            "Some of the important hyperparameters is not set in the program.",
            "hyperparameters-scikitlearn",
            "For learning algorithms, hyperparameters should be tuned and set.",
        ),
    }

    options = (
        (
            "strict_hyperparameters_scikitlearn",
            {
                "default": False,
                "type": "yn",
                "metavar": "<y_or_n>",
                "help": "Force that all parameters of learning algorithms are set.",
            },
        ),
    )

    def __init__(self, linter: PyLinter = HyperparameterChecker):
        super().__init__(linter)
        self.HYPERPARAMETER_RESOURCE = "hyperparameters_scikitlearn_dict.pickle"
        self.MESSAGE = "hyperparameters-scikitlearn"
        self.LIBRARY = "scikitlearn"
        # Main hyperparameters of learning algorithms, as defined in research.
        # Sources:
        # 1. Probst, P., Boulesteix, A. L., & Bischl, B. (2019). Tunability: Importance of
        #   Hyperparameters of Machine Learning Algorithms. Journal of Machine Learning Research,
        #   20(53), 1-32.
        # 2. van Rijn, J. N., & Hutter, F. (2018, July). Hyperparameter importance across datasets.
        #   In Proceedings of the 24th ACM SIGKDD International Conference on Knowledge Discovery &
        #   Data Mining (pp. 2367-2376).
        self.HYPERPARAMETERS_MAIN = {
            # sklearn.ensemble
            "AdaBoostClassifier": {"positional": 5, "keywords": ["learning_rate"]},
            "AdaBoostRegressor": {"positional": 5, "keywords": ["learning_rate"]},
            "GradientBoostingClassifier": {"positional": 20, "keywords": ["learning_rate"]},
            "GradientBoostingRegressor": {"positional": 21, "keywords": ["learning_rate"]},
            "HistGradientBoostingClassifier": {"positional": 18, "keywords": ["learning_rate"]},
            "HistGradientBoostingRegressor": {"positional": 18, "keywords": ["learning_rate"]},
            "RandomForestClassifier": {"positional": 18, "keywords": ["min_samples_leaf", "max_features"],},
            "RandomForestRegressor": {"positional": 17, "keywords": ["min_samples_leaf", "max_features"],},
            # sklearn.linear_model
            "ElasticNet": {"positional": 12, "keywords": ["alpha", "l1_ratio"]},
            # sklearn.neighbors
            "NearestNeighbors": {"positional": 8, "keywords": ["n_neighbors"]},
            # sklearn.svm
            "NuSVC": {"positional": 15, "keywords": ["nu", "kernel", "gamma"]},
            "NuSVR": {"positional": 11, "keywords": ["C", "kernel", "gamma"]},
            "SVC": {"positional": 15, "keywords": ["C", "kernel", "gamma"]},
            "SVR": {"positional": 11, "keywords": ["C", "kernel", "gamma"]},
            # sklearn.tree
            "DecisionTreeClassifier": {"positional": 12, "keywords": ["ccp_alpha"]},
            "DecisionTreeRegressor": {"positional": 11, "keywords": ["ccp_alpha"]},
        }



"""Class which tests TensorArrayTensorflowChecker."""
import astroid as astroid
import pylint.testutils

import dslinter


class TestTensorArrayTensorflowChecker(pylint.testutils.CheckerTestCase):
    """Class which tests TensorArrayTensorflowChecker."""

    CHECKER_CLASS = dslinter.plugin.TensorArrayTensorflowChecker

    def test_not_use_tensor_array(self):
        """Message should be added if tf.constant() type variable is changed in the loop."""
        script = """
        import tensorflow as tf
        @tf.function
        def fibonacci(n):
            a = tf.constant(1)
            b = tf.constant(1)
            c = tf.constant([1, 1])
        
            for i in range(2, n): #@
                a, b = b, a + b
                c = tf.concat([c, [b]], 0)
            
            return c
            
        n = tf.constant(5)
        d = fibonacci(n)        
        """
        module = astroid.parse(script)
        for_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="tensor-array-tensorflow", node=for_node)):
            self.checker.visit_module(module)
            self.checker.visit_for(for_node)

    def test_use_tensor_array(self):
        """No message should be added if tf.TensorArray() type variable is changed in the loop."""
        script = """
        import tensorflow as tf
        @tf.function
        def fibonacci(n):
            a = tf.constant(1)
            b = tf.constant(1)
            c = tf.TensorArray(tf.int32, n)
            c = c.write(0, a)
            c = c.write(1, b)
        
            for i in range(2, n): #@
                a, b = b, a + b
                c = c.write(i, b)
            
            return c.stack()
            
        n = tf.constant(5)
        d = fibonacci(n)        
        """
        module = astroid.parse(script)
        for_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)
            self.checker.visit_for(for_node)



"""Class which tests the ImportChecker."""
import astroid
import pylint.testutils
import dslinter


class TestImportChecker(pylint.testutils.CheckerTestCase):
    """Class which tests the ImportChecker."""

    CHECKER_CLASS = dslinter.plugin.ImportChecker

    def test_finds_incorrect_pandas(self):
        """Test if message is added when pandas is imported incorrectly."""
        script = """
            import pandas as pand #@
            """
        import_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="import-pandas", node=import_node),):
            self.checker.visit_import(import_node)

    def test_ignore_correct_pandas(self):
        """Test if no message is added when pandas is imported correctly."""
        script = """
            import pandas as pd #@
            """
        import_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_import(import_node)

    def test_finds_incorrect_numpy(self):
        """Test if message is added when numpy is imported incorrectly."""
        script = """
            import numpy #@
            """
        import_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="import-numpy", node=import_node),):
            self.checker.visit_import(import_node)

    def test_ignore_correct_numpy(self):
        """Test if no message is added when numpy is imported correctly."""
        script = """
            import numpy as np #@
            """
        import_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_import(import_node)

    def test_finds_incorrect_pyplot(self):
        """Test if message is added when pyplot is imported incorrectly."""
        script = """
            import matplotlib.pyplot as plot #@
            """
        import_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="import-pyplot", node=import_node),):
            self.checker.visit_import(import_node)

    def test_ignore_correct_pyplot(self):
        """Test if no message is added when pyplot is imported correctly."""
        script = """
            import matplotlib.pyplot as plt #@
            """
        import_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_import(import_node)

    def test_finds_incorrect_sklearn(self):
        """Test if message is added when an import from the sklearn module is assigned an alias."""
        script = """
            from sklearn.cluster import KMeans as km #@
            """
        import_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="import-sklearn", node=import_node),):
            self.checker.visit_importfrom(import_node)

    def test_ignore_correct_sklearn(self):
        """Test if no message is added when an sklearn module import is not assigned an alias."""
        script = """
            from sklearn.cluster import KMeans #@
            """
        import_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_importfrom(import_node)

    def test_finds_incorrect_tensorflow(self):
        """Test if message is added when tensorflow is imported incorrectly."""
        script = """
            import tensorflow as tff
            """
        import_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="import-tensorflow", node=import_node)):
            self.checker.visit_import(import_node)

    def test_ignore_correct_tensorflow(self):
        """Test if no message is added when tensorflow is imported correctly."""
        script = """
            import tensorflow as tf
            """
        import_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_import(import_node)

    def test_finds_incorrect_pytorch(self):
        """Test if message is added when pytorch is imported incorrectly."""
        script = """
            import torch as tr
            """
        import_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="import-pytorch", node=import_node)):
            self.checker.visit_import(import_node)

    def test_ignore_correct_pytorch(self):
        """Test if no message is added when pytorch is imported correclty"""
        script =  """
            import torch
            """
        import_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_import(import_node)



"""Checker that checks whether datatype is set when dataframe is imported from data."""
import astroid as astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker

from dslinter.utils.exception_handler import ExceptionHandler


class DatatypePandasChecker(BaseChecker):
    """Checker that checks whether datatype is set when a dataframe is imported from data."""

    __implements__ = IAstroidChecker

    name = "datatype-pandas"
    priority = -1
    msgs = {
        "R5503":(
            "Datatype is not set when a dataframe is imported from data.",
            "datatype-pandas",
            "Datatype should be set when a dataframe is imported from data.",
        )
    }
    options = ()

    _data_import_functions = ["read_csv", "read_table", "read_excel"]

    def visit_call(self, call_node: astroid.Call):
        """
        Vist call node and see whether datatype is set when a dataframe is imported from data.
        :param call_node:
        :return:
        """
        try:
            if(
                hasattr(call_node.func, "attrname")
                and call_node.func.attrname in self._data_import_functions
                and hasattr(call_node.func, "expr")
                and hasattr(call_node.func.expr, "name")
                and call_node.func.expr.name in ["pandas", "pd"]
            ):
                kws = [kw.arg for kw in call_node.keywords if hasattr(kw, "arg")]
                if "dtype" not in kws:
                    self.add_message(msgid="datatype-pandas", node=call_node)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, call_node)



"""Main module for the plugin."""
from dslinter.checkers.chain_indexing_pandas import ChainIndexingPandasChecker
from dslinter.checkers.column_selection_pandas import ColumnSelectionPandasChecker
from dslinter.checkers.dataframe_conversion_pandas import DataframeConversionPandasChecker
from dslinter.checkers.datatype_pandas import DatatypePandasChecker
from dslinter.checkers.dependent_threshold_pytorch import DependentThresholdPytorchChecker
from dslinter.checkers.dependent_threshold_scikitlearn import DependentThresholdScikitLearnChecker
from dslinter.checkers.dependent_threshold_tensorflow import DependentThresholdTensorflowChecker
from dslinter.checkers.forward_pytorch import ForwardPytorchChecker
from dslinter.checkers.gradient_clear_pytorch import GradientClearPytorchChecker
from dslinter.checkers.imports import ImportChecker
from dslinter.checkers.inplace_pandas import InPlacePandasChecker
from dslinter.checkers.deprecated.inplace_numpy import InPlaceNumpyChecker
from dslinter.checkers.mask_missing_pytorch import MaskMissingPytorchChecker
from dslinter.checkers.mask_missing_tensorflow import MaskMissingTensorflowChecker
from dslinter.checkers.memory_release_tensorflow import MemoryReleaseTensorflowChecker
from dslinter.checkers.merge_parameter_pandas import MergeParameterPandasChecker
from dslinter.checkers.deprecated.mode_toggling_pytorch import ModeTogglingPytorchChecker
from dslinter.checkers.randomness_control_numpy import RandomnessControlNumpyChecker
from dslinter.checkers.randomness_control_pytorch import RandomnessControlPytorchChecker
# pylint: disable = line-too-long
from dslinter.checkers.randomness_control_dataloader_pytorch import RandomnessControlDataloaderPytorchChecker
# pylint: disable = line-too-long
from dslinter.checkers.randomness_control_tensorflow import RandomnessControlTensorflowChecker
from dslinter.checkers.randomness_control_scikitlearn import RandomnessControlScikitLLearnChecker
from dslinter.checkers.tensor_array_tensorflow import TensorArrayTensorflowChecker
from dslinter.checkers.unnecessary_iteration_pandas import UnnecessaryIterationPandasChecker
from dslinter.checkers.unnecessary_iteration_tensorflow import UnnecessaryIterationTensorflowChecker
from dslinter.checkers.deterministic_pytorch import DeterministicAlgorithmChecker
from dslinter.checkers.data_leakage_scikitlearn import DataLeakageScikitLearnChecker
from dslinter.checkers.hyperparameters_pytorch import HyperparameterPyTorchChecker
from dslinter.checkers.hyperparameters_tensorflow import HyperparameterTensorflowChecker
# pylint: disable = line-too-long
from dslinter.checkers.hyperparameters_scikitlearn import HyperparameterScikitLearnChecker
from dslinter.checkers.nan_numpy import NanNumpyChecker
from dslinter.checkers.scaler_missing_scikitlearn import ScalerMissingScikitLearnChecker


def register(linter):
    """
    Register checkers of the plugin with the linter.

    :param linter: Linter to add the checkers to.
    """
    linter.register_checker(ImportChecker(linter))
    linter.register_checker(InPlacePandasChecker(linter))
    linter.register_checker(InPlaceNumpyChecker(linter))
    linter.register_checker(UnnecessaryIterationPandasChecker(linter))
    linter.register_checker(UnnecessaryIterationTensorflowChecker(linter))
    linter.register_checker(ScalerMissingScikitLearnChecker(linter))
    linter.register_checker(HyperparameterPyTorchChecker(linter))
    linter.register_checker(HyperparameterTensorflowChecker(linter))
    linter.register_checker(HyperparameterScikitLearnChecker(linter))
    linter.register_checker(MemoryReleaseTensorflowChecker(linter))
    linter.register_checker(DeterministicAlgorithmChecker(linter))
    linter.register_checker(RandomnessControlScikitLLearnChecker(linter))
    linter.register_checker(RandomnessControlPytorchChecker(linter))
    linter.register_checker(RandomnessControlDataloaderPytorchChecker(linter))
    linter.register_checker(RandomnessControlTensorflowChecker(linter))
    linter.register_checker(RandomnessControlNumpyChecker(linter))
    linter.register_checker(DataLeakageScikitLearnChecker(linter))
    linter.register_checker(DependentThresholdPytorchChecker(linter))
    linter.register_checker(DependentThresholdTensorflowChecker(linter))
    linter.register_checker(DependentThresholdScikitLearnChecker(linter))
    linter.register_checker(MaskMissingTensorflowChecker(linter))
    linter.register_checker(MaskMissingPytorchChecker(linter))

    linter.register_checker(NanNumpyChecker(linter))
    linter.register_checker(ChainIndexingPandasChecker(linter))
    linter.register_checker(MergeParameterPandasChecker(linter))
    linter.register_checker(DatatypePandasChecker(linter))
    linter.register_checker(ColumnSelectionPandasChecker(linter))
    linter.register_checker(DataframeConversionPandasChecker(linter))
    linter.register_checker(TensorArrayTensorflowChecker(linter))
    linter.register_checker(ForwardPytorchChecker(linter))
    linter.register_checker(ModeTogglingPytorchChecker(linter))
    linter.register_checker(GradientClearPytorchChecker(linter))



"""Class which tests RandomnessControllingDataLoaderPytorchChecker"""
import astroid
import pylint.testutils
import dslinter


class TestRandomnessControlDataloaderPytorchChecker(pylint.testutils.CheckerTestCase):
    """Class which tests RandomnessControllingDataloaderPytorchChecker"""

    CHECKER_CLASS = dslinter.plugin.RandomnessControlDataloaderPytorchChecker

    def test_with_dataloader_randomness_control1(self):
        """Tests whether no message is added if manual seed is set."""
        script = """
            import torch
            import numpy
            import random
            from torch.utils.data import DataLoader #@

            def seed_worker(worker_id):
                worker_seed = torch.initial_seed() % 2**32
                numpy.random.seed(worker_seed)
                random.seed(worker_seed)

            g = torch.Generator()
            g.manual_seed(0)

            batch_size = 4
            num_workers = 4
            train_dataset = [1, 1, 1, 1]

            train_dataloader = DataLoader( #@
                train_dataset,
                batch_size=batch_size,
                num_workers=num_workers,
                worker_init_fn=seed_worker,
                generator=g,
            )
        """
        import_from_node, assign_node = astroid.extract_node(script)
        call_node = assign_node.value
        with self.assertNoMessages():
            self.checker.visit_importfrom(import_from_node)
            self.checker.visit_call(call_node)

    def test_without_dataloader_randomness_control1(self):
        """Tests whether a message is added if manual seed is not set"""
        script = """
            from torch.utils.data import DataLoader #@
            train_dataloader = DataLoader( #@
                train_dataset,
                batch_size=batch_size,
                num_workers=num_workers
            )
        """
        import_from_node, assign_node = astroid.extract_node(script)
        call_node = assign_node.value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="randomness-control-dataloader-pytorch", node=call_node)):
            self.checker.visit_importfrom(import_from_node)
            self.checker.visit_call(call_node)

    def test_with_dataloader_randomness_control2(self):
        """Tests whether no message is added if manual seed is set."""
        script = """
            import torch
            import numpy
            import random

            def seed_worker(worker_id):
                worker_seed = torch.initial_seed() % 2**32
                numpy.random.seed(worker_seed)
                random.seed(worker_seed)

            g = torch.Generator()
            g.manual_seed(0)

            batch_size = 4
            num_workers = 4
            train_dataset = [1, 1, 1, 1]

            train_dataloader = torch.utils.data.DataLoader( #@
                train_dataset,
                batch_size=batch_size,
                num_workers=num_workers,
                worker_init_fn=seed_worker,
                generator=g,
            )
        """
        call_node = astroid.extract_node(script).value
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    def test_without_dataloader_randomness_control2(self):
        """Tests whether a message is added if manual seed is not set"""
        script = """
            from torch.utils.data import DataLoader
            train_dataloader = torch.utils.data.DataLoader( #@
                train_dataset,
                batch_size=batch_size,
                num_workers=num_workers
            )
        """
        call_node = astroid.extract_node(script).value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="randomness-control-dataloader-pytorch", node=call_node)):
            self.checker.visit_call(call_node)



import astroid
import dslinter.plugin
import pylint.testutils


class TestScalerMissingScikitLearnChecker(pylint.testutils.CheckerTestCase):
    """Class which tests ScalerMissingScikitLearnChecker"""

    CHECKER_CLASS = dslinter.plugin.ScalerMissingScikitLearnChecker

    def test_PCA_missing_scaling(self):
        assign_node = astroid.extract_node("""
            from sklearn.datasets import load_wine
            from sklearn.model_selection import train_test_split
            from sklearn.pipeline import make_pipeline
            from sklearn.decomposition import PCA
            from sklearn.naive_bayes import GaussianNB
            from sklearn.metrics import accuracy_score

            RANDOM_STATE = 42
            features, target = load_wine(return_X_y=True)
            X_train, X_test, y_train, y_test = train_test_split(
                features, target, test_size=0.30, random_state=RANDOM_STATE
            )
            clf = make_pipeline(PCA(n_components=2), GaussianNB()) #@
            clf.fit(X_train, y_train)
            pred_test = clf.predict(X_test)
            ac = accuracy_score(y_test, pred_test)
        """)
        call_node = assign_node.value
        with self.assertAddsMessages(
            pylint.testutils.MessageTest(
                msg_id= 'scaler-missing-scikitlearn',
                node = call_node,
            )
        ):
            self.checker.visit_call(call_node)

    def test_PCA_with_scaling(self):
        assign_node = astroid.extract_node("""
            from sklearn.datasets import load_wine
            from sklearn.model_selection import train_test_split
            from sklearn.pipeline import make_pipeline
            from sklearn.decomposition import PCA
            from sklearn.naive_bayes import GaussianNB
            from sklearn.metrics import accuracy_score
            from sklearn.preprocessing import StandardScaler

            RANDOM_STATE = 42
            features, target = load_wine(return_X_y=True)
            X_train, X_test, y_train, y_test = train_test_split(
                features, target, test_size=0.30, random_state=RANDOM_STATE
            )
            clf = make_pipeline(StandardScaler(), PCA(n_components=2), GaussianNB()) #@
            clf.fit(X_train, y_train)
            pred_test = clf.predict(X_test)
            ac = accuracy_score(y_test, pred_test)
        """)
        call_node = assign_node.value
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    def test_SVC_missing_scaling(self):
        call_node = astroid.extract_node("""
            from sklearn.datasets import load_wine
            from sklearn.model_selection import train_test_split
            from sklearn.metrics import accuracy_score
            from sklearn.svm import SVC
            
            # Make a train/tests split using 30% tests size
            RANDOM_STATE = 42
            features, target = load_wine(return_X_y=True)
            X_train, X_test, y_train, y_test = train_test_split(
                features, target, test_size=0.30, random_state=RANDOM_STATE
            )
            
            # Fit to data and predict using pipelined GNB and PCA
            clf = SVC() 
            clf.fit(X_train, y_train) #@
            pred_test = clf.predict(X_test)
            ac = accuracy_score(y_test, pred_test)
        """)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id='scaler-missing-scikitlearn', node = call_node)):
            self.checker.visit_call(call_node)


    def test_SVC_with_scaling(self):
        assign_node = astroid.extract_node("""
            from sklearn.datasets import load_wine
            from sklearn.model_selection import train_test_split
            from sklearn.metrics import accuracy_score
            from sklearn.svm import SVC
            from sklearn.pipeline import make_pipeline
            from sklearn.preprocessing import StandardScaler
            
            # Make a train/tests split using 30% tests size
            RANDOM_STATE = 42
            features, target = load_wine(return_X_y=True)
            X_train, X_test, y_train, y_test = train_test_split(
                features, target, test_size=0.30, random_state=RANDOM_STATE
            )
            
            # Fit to data and predict using pipelined GNB and PCA
            clf = make_pipeline(StandardScaler(), SVC()) #@
            clf.fit(X_train, y_train)
            pred_test = clf.predict(X_test)
            ac = accuracy_score(y_test, pred_test)
        """)
        call_node = assign_node.value
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    def test_SVC_with_scaling2(self):
        call_node = astroid.extract_node("""
            from sklearn.datasets import load_wine
            from sklearn.model_selection import train_test_split
            from sklearn.metrics import accuracy_score
            from sklearn.svm import SVC
            from sklearn.preprocessing import StandardScaler
            
            # Make a train/tests split using 30% tests size
            RANDOM_STATE = 42
            features, target = load_wine(return_X_y=True)
            X_train, X_test, y_train, y_test = train_test_split(
                features, target, test_size=0.30, random_state=RANDOM_STATE
            )
            scaler = StandardScaler()
            X_train = scaler.fit_transform(X_train)
            X_test = scaler.fit_transform(X_test)
            
            # Fit to data and predict using pipelined GNB and PCA
            clf = SVC() 
            clf.fit(X_train, y_train) #@
            pred_test = clf.predict(X_test)
            ac = accuracy_score(y_test, pred_test)
        """)
        with self.assertNoMessages():
            self.checker.visit_call(call_node)



"""Utility module for type inference."""
import os
from typing import Callable, Dict, List, Tuple

import astroid
import mypy.api

from dslinter.utils.ast import ASTUtil
import uuid


class TypeInference:
    """Utility class for type inference."""

    #pylint: disable = line-too-long
    @staticmethod
    def infer_types(module: astroid.Module, node_type: type, expr: Callable) -> Dict[astroid.node_classes.NodeNG, str]:
        """
        Infer the types of an attribute of all nodes of the same type in a module.

        :param module: The module node where all nodes are located in.
        :param node_type: Type of node of which the type will be inferred on a certain attribute.
        :param expr: Expression to extract the attribute from the node where the type will be
            inferred on. E.g., lambda node: node.func.expr.name
        :return: All nodes in the module of type 'node_type' with the inferred type of the attribute
            accessible with the expression 'expr'.
        """
        nodes = ASTUtil.search_nodes(module, node_type)
        source_code = ASTUtil.get_source_code(module)
        mypy_code = TypeInference.add_reveal_type_calls(source_code, nodes, expr)
        mypy_result = TypeInference.run_mypy(mypy_code)
        try:
            mypy_types = TypeInference.parse_mypy_result(mypy_result)
        except SyntaxError as ex:
            mypy_code_split = mypy_code.splitlines()
            faulty_code = mypy_code_split[int(ex.lineno) - 1]
            if "; reveal_type(" in faulty_code:
                original_code = faulty_code.split("; reveal_type(")[0]
                mypy_code_split[int(ex.lineno) - 1] = original_code
                mypy_types = TypeInference.parse_mypy_result("\n".join(mypy_code_split))
                return TypeInference.combine_nodes_with_inferred_types(nodes, mypy_types)
            else:
                print(
                    "Skipping type checking of module {}: {}. Line {}: {}".format(
                        module.name, ex.msg, ex.lineno, faulty_code
                    )
                )
                return {}
        return TypeInference.combine_nodes_with_inferred_types(nodes, mypy_types)

    @staticmethod
    def add_reveal_type_calls(code: str, nodes: List, expr: Callable) -> str:
        """
        Add reveal_type() calls to source code, so mypy will infer the types.

        :param code: Code to add calls to.
        :param nodes: Nodes of which the type will be inferred on a certain attribute.
        :param expr: Expression to extract the attribute from the node where the type will be
            inferred on. E.g., lambda node: node.func.expr.name
        :return: Code including the calls.
        """
        lines = code.splitlines()
        for node in nodes:
            try:
                line_no = TypeInference.line_to_add_call(node) - 1
                lines[line_no] += "; reveal_type({})".format(expr(node))
                if lines[line_no].strip()[0] == ";":
                    # If the call is added to a new line, remove the semicolon.
                    lines[line_no] = lines[line_no].strip()[1:]
            except AttributeError:
                pass  # The attribute from the expression is not found. Continue.

        return "\n".join(lines)

    @staticmethod
    def line_to_add_call(node: astroid.node_classes.NodeNG):
        """
        Determine the line number where the reveal_type() call can be added to.

        In case the node starts a block, the call cannot be added to the end of the line, as it
        will break the syntax. An example how this would break the syntax:
        '''
        for x in y:;reveal_type(x)
            print(x)
        '''
        The call will be added to the end of the first non block node in the body of the node:
        '''
        for x in y:
            for z in x:
                print(z);reveal_type(x)
        '''

        :param node: The node where a call is added to in the source code.
        :return: Line number where the reveal_type() call can be added.
        """
        if hasattr(node.parent, "blockstart_tolineno") and node not in node.parent.body:
            return TypeInference.line_to_add_call(node.parent.body[0])
        return node.tolineno

    @staticmethod
    def run_mypy(code: str) -> str:
        """
        Run mypy on some code.

        :param code: Code to run mypy on.
        :return: Normal report written to sys.stdout by mypy.
        """
        id1 = str(uuid.uuid1())
        _tmp_dslinter_file = "_tmp_dslinter"+id1+".py"
        file = open(_tmp_dslinter_file, "w", encoding="utf-8")
        file.write(code)
        file.close()
        result = mypy.api.run([_tmp_dslinter_file])
        os.remove(_tmp_dslinter_file)

        if result[1] != '':
            raise Exception("Running mypy resulted in an error: " + result[1])
        return result[0]

    @staticmethod
    def parse_mypy_result(mypy_result: str) -> List[Tuple[int, str]]:
        """
        Parse the result of mypy to obtain all revealed types.

        :param mypy_result: mypy result to parse.
        :return: List of (line number, inferred type) Tuples.
        """
        if "error: invalid syntax" in mypy_result:
            lineno = mypy_result.split(":")[1]
            raise SyntaxError("Mypy cannot run on invalid syntax.", ("", lineno, 0, ""))

        types = []
        revealed_type_indicator = ': note: Revealed type is '
        for line in mypy_result.splitlines():
            if revealed_type_indicator in line:
                line_number = int(line.split(revealed_type_indicator)[0].split(":")[-1])
                inferred_type = line.split(revealed_type_indicator)[1]
                types.append((line_number, inferred_type))
        return types

    @staticmethod
    def combine_nodes_with_inferred_types(
        nodes: List[astroid.node_classes.NodeNG], types: List[Tuple[int, str]]
    ) -> Dict[astroid.node_classes.NodeNG, str]:
        """
        Create a Dict with nodes and their inferred types.

        :param nodes: Nodes where a type is inferred from.
        :param types: List of (line number, inferred type) Tuples.
        :return: Dict with nodes and their inferred types.
        """
        unseen_types = types.copy()
        nodes_with_types = {}
        for node in nodes:
            for line, type_inferred in unseen_types:
                if TypeInference.line_to_add_call(node) == line:
                    nodes_with_types[node] = type_inferred
                    # Remove the tuple so multiple calls on the same line get the correct type.
                    unseen_types.remove((line, type_inferred))
        return nodes_with_types

    @staticmethod
    def infer_variable_most_recent_full_types(module: astroid.Module) -> Dict[str, str]:
        """
        When there is no stub available for a library (e.g., missing tensorflow-stubs),
        use this method instead of infer_types. Infer variable type in Assign nodes.
        :param module: code module
        :return: Dict witn variable names and their inferred type
        """
        variables_with_full_types = {}
        nodes = ASTUtil.search_nodes(module, astroid.Assign)
        for node in nodes:
            if hasattr(node, "targets"):
                for var in node.targets:
                    if hasattr(var, "name"):
                        variable_name = var.name
                        if isinstance(node.value, astroid.nodes.Const):
                            variables_with_full_types[variable_name] = 'const'
                        elif hasattr(node.value, "func"):
                            call = node.value.func
                            full_type = ""
                            if hasattr(call, "attrname"):
                                full_type = "." + call.attrname + full_type
                            while hasattr(call, "expr"):
                                call = call.expr
                                if hasattr(call, "attrname"):
                                    full_type = "." + call.attrname + full_type
                            if hasattr(call, "name"):
                                full_type = call.name + full_type
                            variables_with_full_types[variable_name] = full_type

        return variables_with_full_types

    @staticmethod
    def infer_variable_full_types(module: astroid.Module) -> Dict[str, str]:
        """
        When there is no stub available for a library (e.g., missing tensorflow-stubs),
        use this method instead of infer_types. Infer variable type in Assign nodes.
        :param module: code module
        :return: Dict witn variable names and their inferred type
        """
        variables_with_full_types = {}
        nodes = ASTUtil.search_nodes(module, astroid.Assign)
        for node in nodes:
            if hasattr(node, "targets"):
                for var in node.targets:
                    if hasattr(var, "name"):
                        variable_name = var.name
                        if isinstance(node.value, astroid.nodes.Const):
                            variables_with_full_types[variable_name] = ['const']
                        elif hasattr(node.value, "func"):
                            call = node.value.func
                            full_type = ""
                            if hasattr(call, "attrname"):
                                full_type = "." + call.attrname + full_type
                            while hasattr(call, "expr"):
                                call = call.expr
                                if hasattr(call, "attrname"):
                                    full_type = "." + call.attrname + full_type
                            if hasattr(call, "name"):
                                full_type = call.name + full_type
                            if variable_name not in variables_with_full_types:
                                variables_with_full_types[variable_name] = [full_type]
                            else:
                                variables_with_full_types[variable_name].append(full_type)

        return variables_with_full_types

    @staticmethod
    def infer_native_variable_most_recent_types(module: astroid.Module) -> Dict[str, str]:
        """
        When there is no stub available for a library (e.g., missing tensorflow-stubs),
        use this method instead of infer_types. Infer variable type in Assign nodes.
        :param module: code module
        :return: Dict witn variable names and their inferred type
        """
        variables_with_types = {}
        variables_with_full_types = TypeInference.infer_variable_most_recent_full_types(module)
        for k, v in variables_with_full_types.items():
            variables_with_types[k] = v.split(".")[0]
        return variables_with_types

    @staticmethod
    def infer_library_variable_most_recent_types(module: astroid.Module) -> Dict[str, str]:
        """
        When there is no stub available for a library (e.g., missing tensorflow-stubs),
        use this method instead of infer_types. Infer variable type in Assign nodes.
        :param module: code module
        :return: Dict witn variable names and their inferred type
        """
        variables_with_types = {}
        variables_with_full_types = TypeInference.infer_variable_most_recent_full_types(module)
        for k, v in variables_with_full_types.items():
            strings = v.split(".")
            if len(strings) >= 2:
                variables_with_types[k] = strings[0] + "." + strings[1]
        return variables_with_types

    @staticmethod
    def infer_library_variable_first_types(module: astroid.Module) -> Dict[str, str]:
        """
        When there is no stub available for a library (e.g., missing tensorflow-stubs),
        use this method instead of infer_types. Infer variable type in Assign nodes.
        :param module: code module
        :return: Dict witn variable names and their inferred type
        """
        variables_with_types = {}
        variables_with_full_types = TypeInference.infer_variable_full_types(module)
        for k, v in variables_with_full_types.items():
            strings = v[0].split(".")
            if len(strings) >= 2:
                variables_with_types[k] = strings[0] + "." + strings[1]
        return variables_with_types



"""Class which tests MaskMissingTensorflowChecker."""
import astroid
import pylint.testutils

import dslinter.plugin


class TestMaskMissingTensorflowChecker(pylint.testutils.CheckerTestCase):
    """Class which tests MaskMissingTensorflowChecker."""

    CHECKER_CLASS = dslinter.plugin.MaskMissingTensorflowChecker

    def test_missing_mask_1(self):
        script = """
        cross_entropy = tf.reduce_mean(-tf.reduce_sum(y_*tf.log(y_conv),reduction_indices=[1])) #@
        """
        module = astroid.parse(script)
        node = astroid.extract_node(script)
        call_node = node.value.args[0].operand.args[0].right
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="missing-mask-tensorflow", node = call_node)):
            self.checker.visit_module(module)
            self.checker.visit_call(call_node)

    def test_missing_mask_2(self):
        script = """
        cross_entropy = tf.reduce_mean(-tf.reduce_sum(y_*tf.log(y_conv + 1e-9),reduction_indices=[1])) #@
        """
        module = astroid.parse(script)
        node = astroid.extract_node(script)
        call_node = node.value.args[0].operand.args[0].right
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="missing-mask-tensorflow", node = call_node)):
            self.checker.visit_module(module)
            self.checker.visit_call(call_node)

    def test_with_mask_1(self):
        script = """
        cross_entropy = tf.reduce_mean(-tf.reduce_sum(y_*tf.log(tf.clip_by_value(y_conv,1e-10,1.0)),reduction_indices=[1])) #@
        """
        module = astroid.parse(script)
        node = astroid.extract_node(script)
        call_node = node.value.args[0].operand.args[0].right
        with self.assertNoMessages():
            self.checker.visit_module(module)
            self.checker.visit_call(call_node)

    def test_with_mask_2(self):
        script = """
        y_conv = tf.clip_by_value(y_conv,1e-10,1.0)
        cross_entropy = tf.reduce_mean(-tf.reduce_sum(y_*tf.log(y_conv),reduction_indices=[1])) #@
        """
        module = astroid.parse(script)
        node = astroid.extract_node(script)
        call_node = node.value.args[0].operand.args[0].right
        with self.assertNoMessages():
            self.checker.visit_module(module)
            self.checker.visit_call(call_node)



"""Get all parameters of the learning algorithms in scikit-learn 0.22.2."""
# pylint: disable = line-too-long
# pylint: disable = import-error
from sklearn.calibration import CalibratedClassifierCV
from sklearn.cluster import AffinityPropagation, AgglomerativeClustering, Birch, DBSCAN, FeatureAgglomeration, KMeans, MiniBatchKMeans, MeanShift, OPTICS, SpectralClustering, SpectralBiclustering, SpectralCoclustering
from sklearn.covariance import EmpiricalCovariance, EllipticEnvelope, GraphicalLasso, GraphicalLassoCV, LedoitWolf, MinCovDet, OAS, ShrunkCovariance
from sklearn.cross_decomposition import CCA, PLSCanonical, PLSRegression, PLSSVD
from sklearn.decomposition import DictionaryLearning, FactorAnalysis, FastICA, IncrementalPCA, KernelPCA, LatentDirichletAllocation, MiniBatchDictionaryLearning, MiniBatchSparsePCA, NMF, PCA, SparsePCA, SparseCoder, TruncatedSVD
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis, QuadraticDiscriminantAnalysis
from sklearn.ensemble import AdaBoostClassifier, AdaBoostRegressor, BaggingClassifier, BaggingRegressor, ExtraTreesClassifier, ExtraTreesRegressor, GradientBoostingClassifier, GradientBoostingRegressor, IsolationForest, RandomForestClassifier, RandomForestRegressor, RandomTreesEmbedding, StackingClassifier, StackingRegressor, VotingClassifier, VotingRegressor, HistGradientBoostingClassifier, HistGradientBoostingRegressor
from sklearn.feature_selection import GenericUnivariateSelect, SelectPercentile, SelectKBest, SelectFpr, SelectFdr, SelectFromModel, SelectFwe, RFE, RFECV, VarianceThreshold
from sklearn.gaussian_process import GaussianProcessClassifier, GaussianProcessRegressor
from sklearn.isotonic import IsotonicRegression
from sklearn.kernel_ridge import KernelRidge
from sklearn.linear_model import LogisticRegression, LogisticRegressionCV, PassiveAggressiveClassifier, Perceptron, RidgeClassifier, RidgeClassifierCV, SGDClassifier, LinearRegression, Ridge, RidgeCV, SGDRegressor, ElasticNet, ElasticNetCV, Lars, LarsCV, Lasso, LassoCV, LassoLars, LassoLarsCV, LassoLarsIC, OrthogonalMatchingPursuit, OrthogonalMatchingPursuitCV, ARDRegression, BayesianRidge, MultiTaskElasticNet, MultiTaskElasticNetCV, MultiTaskLasso, MultiTaskLassoCV, HuberRegressor, RANSACRegressor, TheilSenRegressor, PassiveAggressiveRegressor
from sklearn.manifold import Isomap, LocallyLinearEmbedding, MDS, SpectralEmbedding, TSNE
from sklearn.mixture import BayesianGaussianMixture, GaussianMixture
from sklearn.multiclass import OneVsRestClassifier, OneVsOneClassifier, OutputCodeClassifier
from sklearn.multioutput import ClassifierChain, MultiOutputRegressor, MultiOutputClassifier, RegressorChain
from sklearn.naive_bayes import BernoulliNB, CategoricalNB, ComplementNB, GaussianNB, MultinomialNB
from sklearn.neighbors import BallTree, DistanceMetric, KDTree, KernelDensity, KNeighborsClassifier, KNeighborsRegressor, KNeighborsTransformer, LocalOutlierFactor, RadiusNeighborsClassifier, RadiusNeighborsRegressor, RadiusNeighborsTransformer, NearestCentroid, NearestNeighbors, NeighborhoodComponentsAnalysis
from sklearn.neural_network import BernoulliRBM, MLPClassifier, MLPRegressor
from sklearn.semi_supervised import LabelPropagation, LabelSpreading
from sklearn.svm import LinearSVC, LinearSVR, NuSVC, NuSVR, OneClassSVM, SVC, SVR
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor, ExtraTreeClassifier, ExtraTreeRegressor

# Collect learning classes. A class is a learning class if it is described in section '1. Supervised learning' or '2. Unsupervised learning' of the scikit-learn user guide.
from dslinter.scripts.hyperparameters import save_hyperparameter

learning_classes = []
learning_classes.extend([CalibratedClassifierCV])  # calibration
learning_classes.extend([AffinityPropagation, AgglomerativeClustering, Birch, DBSCAN, FeatureAgglomeration, KMeans, MiniBatchKMeans, MeanShift, OPTICS, SpectralClustering, SpectralBiclustering, SpectralCoclustering])  # cluster
learning_classes.extend([EmpiricalCovariance, EllipticEnvelope, GraphicalLasso, GraphicalLassoCV, LedoitWolf, MinCovDet, OAS, ShrunkCovariance])  # covariance
learning_classes.extend([CCA, PLSCanonical, PLSRegression, PLSSVD])  # cross_decomposition
learning_classes.extend([DictionaryLearning, FactorAnalysis, FastICA, IncrementalPCA, KernelPCA, LatentDirichletAllocation, MiniBatchDictionaryLearning, MiniBatchSparsePCA, NMF, PCA, SparsePCA, SparseCoder, TruncatedSVD])  # decomposition
learning_classes.extend([LinearDiscriminantAnalysis, QuadraticDiscriminantAnalysis])  # discriminant_analysis
learning_classes.extend([AdaBoostClassifier, AdaBoostRegressor, BaggingClassifier, BaggingRegressor, ExtraTreesClassifier, ExtraTreesRegressor, GradientBoostingClassifier, GradientBoostingRegressor, IsolationForest, RandomForestClassifier, RandomForestRegressor, RandomTreesEmbedding, StackingClassifier, StackingRegressor, VotingClassifier, VotingRegressor, HistGradientBoostingClassifier, HistGradientBoostingRegressor])  # ensemble
learning_classes.extend([GenericUnivariateSelect, SelectPercentile, SelectKBest, SelectFpr, SelectFdr, SelectFromModel, SelectFwe, RFE, RFECV, VarianceThreshold])  # feature_selection
learning_classes.extend([GaussianProcessClassifier, GaussianProcessRegressor])  # gaussian_process
learning_classes.extend([IsotonicRegression])  # isotonic
learning_classes.extend([KernelRidge])  # kernel_ridge
learning_classes.extend([LogisticRegression, LogisticRegressionCV, PassiveAggressiveClassifier, Perceptron, RidgeClassifier, RidgeClassifierCV, SGDClassifier, LinearRegression, Ridge, RidgeCV, SGDRegressor, ElasticNet, ElasticNetCV, Lars, LarsCV, Lasso, LassoCV, LassoLars, LassoLarsCV, LassoLarsIC, OrthogonalMatchingPursuit, OrthogonalMatchingPursuitCV, ARDRegression, BayesianRidge, MultiTaskElasticNet, MultiTaskElasticNetCV, MultiTaskLasso, MultiTaskLassoCV, HuberRegressor, RANSACRegressor, TheilSenRegressor, PassiveAggressiveRegressor])  # linear_model
learning_classes.extend([Isomap, LocallyLinearEmbedding, MDS, SpectralEmbedding, TSNE])  # manifold
learning_classes.extend([BayesianGaussianMixture, GaussianMixture])  # mixture_classes
learning_classes.extend([OneVsRestClassifier, OneVsOneClassifier, OutputCodeClassifier])  # multiclass
learning_classes.extend([ClassifierChain, MultiOutputRegressor, MultiOutputClassifier, RegressorChain])  # multioutput
learning_classes.extend([BernoulliNB, CategoricalNB, ComplementNB, GaussianNB, MultinomialNB])  # naive_bayes
learning_classes.extend([BallTree, DistanceMetric, KDTree, KernelDensity, KNeighborsClassifier, KNeighborsRegressor, KNeighborsTransformer, LocalOutlierFactor, RadiusNeighborsClassifier, RadiusNeighborsRegressor, RadiusNeighborsTransformer, NearestCentroid, NearestNeighbors, NeighborhoodComponentsAnalysis])  # neighbors
learning_classes.extend([BernoulliRBM, MLPClassifier, MLPRegressor])  # neural_network
learning_classes.extend([LabelPropagation, LabelSpreading])  # semi_supervised
learning_classes.extend([LinearSVC, LinearSVR, NuSVC, NuSVR, OneClassSVM, SVC, SVR])  # svm_classes
learning_classes.extend([DecisionTreeClassifier, DecisionTreeRegressor, ExtraTreeClassifier, ExtraTreeRegressor])  # tree

if __name__ == "__main__":
    save_hyperparameter(learning_classes, "../resources/hyperparameters_scikitlearn_dict.pickle")



"""DataFrame checker which checks correct handling of calls on DataFrames."""
from typing import Dict
import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker
from dslinter.utils.exception_handler import ExceptionHandler
from dslinter.utils.inplace_helper import inplace_is_true
from dslinter.utils.type_inference import TypeInference


class InPlacePandasChecker(BaseChecker):
    """DataFrame checker which checks correct handling of calls on DataFrames."""

    __implements__ = IAstroidChecker

    name = "inplace-pandas"
    priority = -1
    msgs = {
        "W5503": (
            "Result of operation on a DataFrame is not assigned.",
            "inplace-pandas",
            "Most operations on a DataFrame return a new DataFrame. These should be assigned to a variable.",
        ),
    }
    options = ()

    # [node, inferred type of object the function is called on]
    _call_types: Dict[astroid.Call, str] = {}

    # Whitelisted functions for which a DataFrame does not have to be assigned.
    WHITELISTED = [
        "all",
        "any",
        "bool",
        "boxplot",
        "corr",
        "corrwith",
        "count",
        "cov",
        "describe",
        "duplicated",
        "equals",
        "first_valid_index",
        "head",
        "hist",
        "idxmax",
        "idxmin",
        "info",
        "insert",
        "items",
        "iteritems",
        "iterrows",
        "itertuples",
        "keys",
        "last_valid_index",
        "mad",
        "max",
        "mean",
        "median",
        "memory_usage",
        "min",
        "mode",
        "nlargest",
        "nsmallest",
        "nunique",
        "plot",
        "profile_report",  # Jupyter notebook specific
        "sample",
        "sem",
        "skew",
        "sum",
        "tail",
        "update",
        "var",
        "to_***",
    ]

    def visit_module(self, module: astroid.Module):
        """
        When an Module node is visited, scan for Call nodes and get type the function is called on.

        :param module: Node which is visited.
        """
        try:
            # noinspection PyTypeChecker
            # pylint: disable = line-too-long
            self._call_types = TypeInference.infer_types(module, astroid.Call, lambda x: x.func.expr.name)
        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, module)

    def visit_call(self, node: astroid.Call):
        """
        When a Call node is visited, add messages if it violated the defined rules.

        :param node: Node which is visited.
        """
        try:
            if (
                self._is_simple_call_node(node)
                and not self._function_whitelisted(node)
                and self._dataframe_is_lost(node)
            ):
                self.add_message("inplace-pandas", node=node)
        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, node)

    @staticmethod
    def _is_simple_call_node(node: astroid.Call) -> bool:
        """
        Evaluate whether the node is a 'simple' call node.

        A 'simple' Call node is a single function call made on an expression.
        E.g., 'a.f()' and not 'f()' or 'a.f().g()' or 'a.f(g())'.

        :param node: Call node to evaluate.
        :return: True when the Call node is considered simple.
        """
        return (
            hasattr(node.func, "expr")  # The call is made on an expression.
            and hasattr(node.func.expr, "name")  # The expr the func is called on is a named thing.
            and not isinstance(node.parent, astroid.Attribute)  # Call is not an attribute.
            and not isinstance(node.parent, astroid.Call)  # Call is not part of another call.
        )

    @staticmethod
    def _function_whitelisted(node: astroid.Call) -> bool:
        """
        Evaluate whether the a Call node is whitelisted for the unassigned-dataframe rule.

        All whitelisted functions do not have to assigned to a variable. Whitelisted functions
        are listed in the WHITELISTED variable and also include all functions starting with "to_".

        :param node: Call node to evaluate.
        :return: True when the function of the call is whitelisted.
        """
        return hasattr(node.func, "attrname") and (
                node.func.attrname in InPlacePandasChecker.WHITELISTED
                or node.func.attrname[:3] == "to_"
        )

    def _dataframe_is_lost(self, node: astroid.Call) -> bool:
        """
        Check whether the call is done on a DataFrame and the result is lost.

        A result is seen as lost if its parent is an Expression and the operation is not done
        inplace.

        :param node: Node which is visited.
        :return: True when the call results in a DataFrame which is lost.
        """
        return (
            node in self._call_types  # Check if the type is inferred of this call.
            and (
                self._call_types[node] == '"pandas.core.frame.DataFrame"'
                or self._call_types[node] == '"pyspark.sql.dataframe.DataFrame"'
            )
            and not inplace_is_true(node, "inplace")
            # If the parent of the Call is an Expression (not an Assignment),
            # it means the DataFrame is lost.
            and isinstance(node.parent, astroid.Expr)
        )



"""Get all parameters of the learning algorithms in pytorch."""
# pylint: disable = line-too-long
# pylint: disable = import-error
from torch.utils.data import DataLoader
from torch.optim import Adadelta, Adagrad, Adam, AdamW, SparseAdam, Adamax, ASGD, LBFGS, NAdam, RAdam, RMSprop, Rprop, SGD
from dslinter.scripts.hyperparameters import save_hyperparameter

learning_classes = []
learning_classes.extend([DataLoader])  # dataloader
learning_classes.extend([Adadelta, Adagrad, Adam, AdamW, SparseAdam, Adamax, ASGD, LBFGS, NAdam, RAdam, RMSprop, Rprop, SGD]) #optimizer

if __name__ == "__main__":
    save_hyperparameter(learning_classes, "../resources/hyperparameters_pytorch_dict.pickle")



"""Class which tests UnnecessaryIterationTensorflowChecker"""
import astroid
import pylint.testutils
import dslinter.plugin


class TestUnnecessaryIterationTensorflowChecker(pylint.testutils.CheckerTestCase):
    """Class which tests UnnecessaryIterationTensorflowChecker"""
    CHECKER_CLASS = dslinter.plugin.UnnecessaryIterationTensorflowChecker

    def test_iteration(self):
        """Test whether there is a message added when there is a iteration with augmented assign."""
        script = """
        import tensorflow as tf
        x = tf.random.uniform([500, 10]) 
        z = tf.zeros([10])
        for i in range(500): #@
            z += x[i]
        """
        module_node = astroid.parse(script)
        for_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="iteration-tensorflow", node = for_node),):
            self.checker.visit_module(module_node)
            self.checker.visit_for(for_node)

    def test_no_iteration(self):
        """Test whether there is no message added when there is no iteration."""
        script = """
        import tensorflow as tf
        x = tf.random.uniform([500, 10])
        z = tf.reduce_sum(x, axis=0) #@
        """
        module_node = astroid.parse(script)
        for_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_module(module_node)
            self.checker.visit_for(for_node)

    def test_iteration_for_not_tf_variable(self):
        script = """
        m = 1
        for i in range(7):
            for j in range(4): #@
                # k=k+1
                k += m
                sns.distplot(X_train_pt_df['V'+str(k)], ax=ax[i][j])
                ax[i][j].set_title('V'+str(k))
        """
        module_node = astroid.parse(script)
        for_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_module(module_node)
            self.checker.visit_for(for_node)

    def test_iteration_for_not_tf_variable_2(self):
        script = """
        k = 1
        for i in range(7):
            for j in range(4): #@
                # k=k+1
                k += 1
                sns.distplot(X_train_pt_df['V'+str(k)], ax=ax[i][j])
                ax[i][j].set_title('V'+str(k))
        """
        module_node = astroid.parse(script)
        for_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_module(module_node)
            self.checker.visit_for(for_node)



"""Class which tests RandomnessControllingPytorchChecker"""
import astroid
import pylint.testutils
import dslinter


class TestRandomnessControlPytorchChecker(pylint.testutils.CheckerTestCase):
    """Class which tests RandomnessControllingPytorchChecker"""

    CHECKER_CLASS = dslinter.plugin.RandomnessControlPytorchChecker

    def test_with_pytorch_randomness_control(self):
        """Tests whether no message is added if manual seed is set."""
        script = """
        import torch #@
        torch.manual_seed(0) 
        torch.randn(10).index_copy(0, torch.tensor([0]), torch.randn(1))
                
        if __name__ == '__main__':
            pass
        """
        module = astroid.parse(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)

    def test_without_pytorch_randomness_control(self):
        """Tests whether a message is added if manual seed is not set"""
        script = """
        import torch #@
        torch.randn(10).index_copy(0, torch.tensor([0]), torch.randn(1)) 
                
        if __name__ == '__main__':
            pass
        """
        module = astroid.parse(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="randomness-control-pytorch", node=module)):
            self.checker.visit_module(module)

    def test_pytorch_randomness_with_main_module(self):
        script = """
        # import statements
        import torch #@
        import torch.nn as nn
        from torch.utils import data
        
        # set flags / seeds
        torch.backends.cudnn.benchmark = True
        np.random.seed(1)
        torch.manual_seed(1)
        torch.cuda.manual_seed(1)
        
        # Start with main code
        if __name__ == '__main__':
            # argparse for additional flags for experiment
            parser = argparse.ArgumentParser(description="Train a network for ...")
            opt = parser.parse_args() 
        """
        module = astroid.parse(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)




"""Class which tests the HyperparameterTensorflowChecker."""
import astroid
import pylint.testutils

import dslinter.plugin


class TestHyperparameterTensorflowChecker(pylint.testutils.CheckerTestCase):
    """Class which tests the HyperparameterTensorflowChecker."""

    CHECKER_CLASS = dslinter.plugin.HyperparameterTensorflowChecker

    def test_batch_size_set(self):
        script = """
            import tensorflow as tf

            class MyModel(tf.keras.Model):

              def __init__(self, *args, **kwargs):
                super(MyModel, self).__init__(*args, **kwargs)
                self.loss_tracker = tf.keras.metrics.Mean(name='loss')

              def compute_loss(self, x, y, y_pred, sample_weight):
                loss = tf.reduce_mean(tf.math.squared_difference(y_pred, y))
                loss += tf.add_n(self.losses)
                self.loss_tracker.update_state(loss)
                return loss

              def reset_metrics(self):
                self.loss_tracker.reset_states()

              @property
              def metrics(self):
                return [self.loss_tracker]

            tensors = tf.random.uniform((10, 10)), tf.random.uniform((10,))
            dataset = tf.data.Dataset.from_tensor_slices(tensors).repeat().batch(1)

            inputs = tf.keras.layers.Input(shape=(10,), name='my_input')
            outputs = tf.keras.layers.Dense(10)(inputs)
            model = MyModel(inputs, outputs)
            model.add_loss(tf.reduce_sum(outputs))

            optimizer = tf.keras.optimizers.SGD()
            model.compile(optimizer, loss='mse', steps_per_execution=10)
            model.fit(dataset, batch_size = 4, epochs=2, steps_per_epoch=10) #@
            print('My custom loss: ', model.loss_tracker.result().numpy())
        """
        call_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    # def test_batch_size_not_set(self):
    #     script = """
    #         import tensorflow as tf
    #
    #         class MyModel(tf.keras.Model):
    #
    #           def __init__(self, *args, **kwargs):
    #             super(MyModel, self).__init__(*args, **kwargs)
    #             self.loss_tracker = tf.keras.metrics.Mean(name='loss')
    #
    #           def compute_loss(self, x, y, y_pred, sample_weight):
    #             loss = tf.reduce_mean(tf.math.squared_difference(y_pred, y))
    #             loss += tf.add_n(self.losses)
    #             self.loss_tracker.update_state(loss)
    #             return loss
    #
    #           def reset_metrics(self):
    #             self.loss_tracker.reset_states()
    #
    #           @property
    #           def metrics(self):
    #             return [self.loss_tracker]
    #
    #         tensors = tf.random.uniform((10, 10)), tf.random.uniform((10,))
    #         dataset = tf.data.Dataset.from_tensor_slices(tensors).repeat().batch(1)
    #
    #         inputs = tf.keras.layers.Input(shape=(10,), name='my_input')
    #         outputs = tf.keras.layers.Dense(10)(inputs)
    #         model = MyModel(inputs, outputs)
    #         model.add_loss(tf.reduce_sum(outputs))
    #
    #         optimizer = tf.keras.optimizers.SGD()
    #         model.compile(optimizer, loss='mse', steps_per_execution=10)
    #         model.fit(dataset, epochs=2, steps_per_epoch=10) #@
    #         print('My custom loss: ', model.loss_tracker.result().numpy())
    #     """
    #     call_node = astroid.extract_node(script)
    #     with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id = "hyperparameter-tensorflow", node = call_node)):
    #         self.checker.visit_call(call_node)

    def test_learning_rate_set(self):
        script = """
        from tf.keras.optimizers import SGD #@
        optimizer = SGD(learning_rate=0.001, momentum = 0) #@
        """
        importfrom_node, assign_node = astroid.extract_node(script)
        call_node = assign_node.value
        with self.assertNoMessages():
            self.checker.visit_importfrom(importfrom_node)
            self.checker.visit_call(call_node)

    def test_learning_rate_not_set(self):
        script = """
        from tf.keras.optimizers import SGD #@
        optimizer = SGD() #@
        """
        importfrom_node, assign_node = astroid.extract_node(script)
        call_node = assign_node.value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id = "hyperparameters-tensorflow", node = call_node)):
            self.checker.visit_importfrom(importfrom_node)
            self.checker.visit_call(call_node)

    def test_sklearn_fit(self):
        script = """
            from sklearn.datasets import load_wine
            from sklearn.model_selection import train_test_split
            from sklearn.pipeline import make_pipeline
            from sklearn.decomposition import PCA
            from sklearn.naive_bayes import GaussianNB
            from sklearn.metrics import accuracy_score

            RANDOM_STATE = 42
            features, target = load_wine(return_X_y=True)
            X_train, X_test, y_train, y_test = train_test_split(
                features, target, test_size=0.30, random_state=RANDOM_STATE
            )
            clf = make_pipeline(PCA(n_components=2), GaussianNB())
            clf.fit(X_train, y_train) #@
            pred_test = clf.predict(X_test)
            ac = accuracy_score(y_test, pred_test)
        """
        call_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_call(call_node)



"""Check whether the memory is freed in time."""
from typing import Dict
import astroid
from pylint.checkers import BaseChecker
from dslinter.utils.exception_handler import ExceptionHandler
from dslinter.utils.type_inference import TypeInference


class MemoryReleaseTensorflowChecker(BaseChecker):
    """Check whether the memory is freed in time."""

    name = "memory-release-tensorflow"
    priority = -1
    msgs = {
        "W5506":{
            "The memory has not freed in time.",
            "memory-release-tensorflow",
            "The `clean_session()` can be used to free memory in the loop."
        }
    }
    options = ()

    MODELS = [
        "Sequential",
        "Model"
    ]

    # [variable name, inferred type of object the function is called on]
    _variable_types: Dict[str, str] = {}

    def visit_module(self, module: astroid.Module):
        """Visit module and infer which library the variables are from. """
        try:
            self._variable_types = TypeInference.infer_native_variable_most_recent_types(module)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, module)

    def visit_for(self, node: astroid.For):
        """Evaluate whether memory is freed in a loop with model creation."""
        try:
            has_clear_session = False
            has_model_creation = False

            for nod in node.body:
                if (
                    hasattr(nod, "value")
                    and hasattr(nod.value, "func")
                    and hasattr(nod.value.func, "attrname")
                    and nod.value.func.attrname == "clear_session"
                ):
                    has_clear_session = True

                if (
                    hasattr(nod, "value")
                    and hasattr(nod.value, "func")
                    and hasattr(nod.value.func, "attrname")
                    and nod.value.func.attrname in self.MODELS
                ):
                    if(
                        hasattr(nod, "targets")
                        and len(nod.targets) > 0
                        and hasattr(nod.targets[0], "name")
                        and self._is_tf_variable(nod.targets[0].name)
                    ):
                        has_model_creation = True

            if(
                has_clear_session is False
                and has_model_creation is True
            ):
                # if there is no clear_session call in the loop
                # while there is a model creation, the rule is violated.
                self.add_message("memory-release-tensorflow", node=node)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, node)

    def _is_tf_variable(self, name) -> bool:
        """
        Check whether the variable with the name is a tensorflow variable
        :param name: name of the variable
        :return: True when meeting the requirements
        """
        return name in self._variable_types and self._variable_types[name] in ["tf", "tensorflow"]



"""Class which tests the DataFrameChecker."""
import astroid
import pylint.testutils
import dslinter


class TestUnnecessaryIterationPandasChecker(pylint.testutils.CheckerTestCase):
    """Class which tests the DataFrameChecker."""

    CHECKER_CLASS = dslinter.plugin.UnnecessaryIterationPandasChecker

    def test_iterating_through_dataframe(self):
        """Test whether a message is added when there is iterated through a DataFrame."""
        script = """
        import pandas as pd
        df = pd.DataFrame()
        for _, row in df.iterrows(): 
            pass
        """
        module_tree = astroid.parse(script)
        call = module_tree.body[-1].iter
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="dataframe-iteration-modification-pandas", node=call),):
            self.checker.visit_module(module_tree)
            self.checker.visit_call(call)

    def test_iterating_and_modifying(self):
        """Test whether a dataframe-iteration-modification violation is correctly found."""
        script = """
        import pandas as pd
        df = pd.DataFrame()
        for _, row in df.iterrows():
            row['a'] = 10
        """
        module_tree = astroid.parse(script)
        for_node = module_tree.body[-1]
        with self.assertAddsMessages(
            pylint.testutils.MessageTest(msg_id="dataframe-iteration-modification-pandas", node=for_node),
        ):
            self.checker.visit_module(module_tree)
            self.checker.visit_for(for_node)

    def test_iterating_and_modifying_nested(self):
        """Test whether a nested dataframe-iteration-modification violation is correctly found."""
        script = """
        import pandas as pd
        df = pd.DataFrame()
        for _, row in df.iterrows():
            if True: row['a'] = 10
        """
        module_tree = astroid.parse(script)
        for_node = module_tree.body[-1]
        with self.assertAddsMessages(
            pylint.testutils.MessageTest(msg_id="dataframe-iteration-modification-pandas", node=for_node),
        ):
            self.checker.visit_module(module_tree)
            self.checker.visit_for(for_node)
